from io import BytesIO
from .stream import BinStream


class WPKWem:
    __slots__ = (
        'id', 'offset', 'size'
    )

    def __init__(self, id=None, offset=None, size=None):
        self.id = id
        self.offset = offset
        self.size = size

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}
    

class WPK:
    __slots__ = ('signature', 'version', 'wems')

    def __init__(self, signature=None, version=None, wems=None):
        self.signature = signature
        self.version = version
        self.wems = wems

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

    def stream(self, path, mode, raw=None):
        if raw != None:
            if raw == True:  # the bool True value
                return BinStream(BytesIO())
            else:
                return BinStream(BytesIO(raw))
        return BinStream(open(path, mode))

    def read(self, path, raw=None):
        with self.stream(path, 'rb', raw) as bs:
            self.signature, = bs.read_s(4)
            if self.signature != 'r3d2':
                raise Exception(
                    f'pyRitoFile: Failed: Read WPK {path}: Wrong signature file: {hex(self.signature)}')
            self.version, = bs.read_u32()
            # read wems offset in wpk
            wem_count, = bs.read_u32()
            self.wems = [WPKWem() for i in range(wem_count)]
            for wem in self.wems:
                wem.offset, = bs.read_u32()
            # remove any wem that has offset 0 
            for wem in self.wems:
                if wem.offset == 0:
                    self.wems.remove(wem)
            wem_count = len(self.wems)
            # now actually read wem info 
            for wem in self.wems:
                bs.seek(wem.offset)
                # update wem.offset as data offset, not wem offset in wpk
                wem.offset, wem.size = bs.read_u32(2)
                wem.id, = bs.read_c_sep_0(bs.read_u32()[0])
                wem.id = int(wem.id.replace('.wem', ''))

    def write(self, path, wem_datas, raw=None):
        with self.stream(path, 'wb', raw) as bs:
            # magic, version
            bs.write_s('r3d2')
            bs.write_u32(1)
            # wems offsets - write later
            wem_count = len(self.wems)
            bs.write_u32(wem_count)
            for i, wem in enumerate(self.wems):
                bs.pad(4)
            # wem infos
            wem_offsets = []
            wem_data_offsets = []
            for i, wem in enumerate(self.wems):
                wem_offsets.append(bs.tell())
                wem.size = len(wem_datas[i])
                wem_data_offsets.append(bs.tell())
                bs.pad(4)
                bs.write_u32(wem.size)
                wem_id = str(wem.id) + '.wem'
                bs.write_u32(len(wem_id))
                bs.write_c_sep_0(wem_id)
            # wem datas
            for i, wem_data in enumerate(wem_datas):
                data_offset = bs.tell()
                bs.seek(wem_data_offsets[i])
                bs.write_u32(data_offset)
                bs.seek(data_offset)
                bs.write(wem_data)
            # wem offsets
            bs.seek(12)
            for wem_offset in wem_offsets:
                bs.write_u32(wem_offset)

            return bs.raw() if raw else None 

        



from io import BytesIO
from .stream import BinStream
from enum import IntEnum


class TEXFormat(IntEnum):
    ETC1 = 1
    ETC2 = 2
    ETC2_EAC = 3
    DXT1 = 10
    DXT1_ = 11
    DXT5 = 12
    RGBA8 = 20

    def __json__(self):
        return self.name


class TEX:
    __slots__ = (
        'signature', 'width', 'height',
        'format', 'unknown1', 'unknown2', 'mipmaps',
        'data'
    )

    def __init__(self, signature=None, width=None, height=None, format=None, unknown1=None, unknown2=None, mipmaps=False, data=None):
        self.signature = signature
        self.width = width
        self.height = height
        self.format = format
        self.unknown1 = unknown1
        self.unknown2 = unknown2
        self.mipmaps = mipmaps
        self.data = data

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

    def stream(self, path, mode, raw=None):
        if raw != None:
            if raw == True:  # the bool True value
                return BinStream(BytesIO())
            else:
                return BinStream(BytesIO(raw))
        return BinStream(open(path, mode))

    def read(self, path, raw=None):
        with self.stream(path, 'rb', raw) as bs:
            # read headers
            self.signature, = bs.read_u32()
            if self.signature != 0x00584554:
                raise Exception(
                    f'pyRitoFile: Failed: Read TEX {path}: Wrong file signature: {hex(self.signature)}')
            self.width, self.height = bs.read_u16(2)
            self.unknown1, self.format, self.unknown2 = bs.read_u8(3)
            self.format = TEXFormat(self.format)
            self.mipmaps, = bs.read_b()
            # read data
            if self.mipmaps and self.format in (TEXFormat.DXT1, TEXFormat.DXT1_, TEXFormat.DXT5, TEXFormat.RGBA8):
                # if mipmaps and supported format
                if self.format in (TEXFormat.DXT1, TEXFormat.DXT1_):
                    block_size = 4
                    bytes_per_block = 8
                elif self.format == TEXFormat.DXT5:
                    block_size = 4
                    bytes_per_block = 16
                else:
                    block_size = 1
                    bytes_per_block = 4
                mipmap_count = 32 - \
                    len(f'{max(self.width, self.height):032b}'.split(
                        '1', 1)[0])
                self.data = []
                for i in reversed(range(mipmap_count)):
                    current_width = max(self.width // (1 << i), 1)
                    current_height = max(self.height // (1 << i), 1)
                    block_width = (current_width +
                                   block_size - 1) // block_size
                    block_height = (current_height +
                                    block_size - 1) // block_size
                    current_size = bytes_per_block * block_width * block_height
                    self.data.append(bs.read(current_size))
            else:
                self.data = [bs.read(-1)]

    def write(self, path, raw=None):
        with self.stream(path, 'wb', raw) as bs:
            # write headers
            bs.write_u32(0x00584554)
            bs.write_u16(self.width, self.height)
            bs.write_u8(1, self.format.value, 0)  # unknown1, format, unknown2
            bs.write_b(self.mipmaps)
            if self.mipmaps and self.format in (TEXFormat.DXT1, TEXFormat.DXT1_, TEXFormat.DXT5, TEXFormat.RGBA8):
                for block_data in self.data:
                    bs.write(block_data)
            else:
                bs.write(self.data[0])



from struct import Struct
from .structs import Vector, Quaternion, Matrix4


class BinStream:
    def __init__(self, f):
        self.stream = f

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    # stream
    def tell(self):
        return self.stream.tell()

    def seek(self, pos, mode=0):
        self.stream.seek(pos, mode)

    def pad(self, length):
        self.stream.seek(length, 1)

    def end(self):
        return_offset = self.stream.tell()
        self.stream.seek(0, 2)
        e = self.stream.tell()
        self.stream.seek(return_offset)
        return e

    def close(self):
        self.stream.close()

    def raw(self):
        self.seek(0)
        return self.stream.read()

    # read

    def read_fmt(self, fmt, fmt_size = None):
        if not fmt_size:
            fmt_size = Struct(fmt).size
        return Struct(fmt).unpack(self.stream.read(fmt_size))

    def read(self, length):
        return self.stream.read(length)

    def read_b(self, count=1):
        return Struct(f'<{count}?').unpack(self.stream.read(count))

    def read_i8(self, count=1):
        return Struct(f'<{count}b').unpack(self.stream.read(count))

    def read_u8(self, count=1):
        return Struct(f'<{count}B').unpack(self.stream.read(count))

    def read_i16(self, count=1):
        return Struct(f'<{count}h').unpack(self.stream.read(count*2))

    def read_u16(self, count=1):
        return Struct(f'<{count}H').unpack(self.stream.read(count*2))

    def read_i32(self, count=1):
        return Struct(f'<{count}i').unpack(self.stream.read(count*4))

    def read_u32(self, count=1):
        return Struct(f'<{count}I').unpack(self.stream.read(count*4))

    def read_i64(self, count=1):
        return Struct(f'<{count}q').unpack(self.stream.read(count*8))

    def read_u64(self, count=1):
        return Struct(f'<{count}Q').unpack(self.stream.read(count*8))

    def read_f32(self, count=1):
        return Struct(f'<{count}f').unpack(self.stream.read(count*4))

    def read_f64(self, count=1):
        return Struct(f'<{count}d').unpack(self.stream.read(count*8))

    def read_ul(self, count=1): # For rst
        return Struct(f'<{count}L').unpack(self.stream.read(count*4))

    def read_ull(self, count=1): # For rst
        return Struct(f'<{count}Q').unpack(self.stream.read(count*8))

    def read_vec2(self, count=1):
        floats = Struct(f'<{count*2}f').unpack(self.stream.read(count*8))
        return [Vector(floats[i], floats[i+1]) for i in range(0, len(floats), 2)]

    def read_vec3(self, count=1):
        floats = Struct(f'<{count*3}f').unpack(self.stream.read(count*12))
        return [Vector(floats[i], floats[i+1], floats[i+2]) for i in range(0, len(floats), 3)]

    def read_vec4(self, count=1):
        floats = Struct(f'<{count*4}f').unpack(self.stream.read(count*16))
        return [Vector(floats[i], floats[i+1], floats[i+2], floats[i+3]) for i in range(0, len(floats), 4)]

    def read_quat(self, count=1):
        floats = Struct(f'<{count*4}f').unpack(self.stream.read(count*16))
        return [Quaternion(floats[i], floats[i+1], floats[i+2], floats[i+3]) for i in range(0, len(floats), 4)]

    def read_mtx4(self):
        return Matrix4(*Struct('16f').unpack(self.stream.read(64))),
        
    def read_s(self, length, encoding='ascii'):
        return self.stream.read(length).decode(encoding),

    def read_s_padded(self, length, encoding='ascii'):
        return bytes(b for b in self.stream.read(length) if b != 0).decode(encoding),

    def read_s_sized16(self, encoding='ascii'):
        return self.stream.read(Struct('H').unpack(self.stream.read(2))[0]).decode(encoding),

    def read_s_sized32(self, encoding='ascii'):
        return self.stream.read(Struct('I').unpack(self.stream.read(4))[0]).decode(encoding),

    def read_c_until0(self):
        s = ''
        while True:
            c = self.stream.read(1)[0]
            if c == 0:
                break
            s += chr(c)
        return s,

    def read_c_sep_0(self, length):
        s = ''
        for i in range(length):
            s += chr(self.stream.read(1)[0])
            self.pad(1)
        return s,

    # write

    def write_fmt(self, fmt, *values):
        self.stream.write(Struct(fmt).pack(*values))

    def write(self, values):
        self.stream.write(values)

    def write_b(self, *values):
        self.stream.write(Struct(f'<{len(values)}?').pack(*values))

    def write_i8(self, *values):
        self.stream.write(Struct(f'<{len(values)}b').pack(*values))

    def write_u8(self, *values):
        self.stream.write(Struct(f'<{len(values)}B').pack(*values))

    def write_i16(self, *values):
        self.stream.write(Struct(f'<{len(values)}h').pack(*values))

    def write_u16(self, *values):
        self.stream.write(Struct(f'<{len(values)}H').pack(*values))

    def write_i32(self, *values):
        self.stream.write(Struct(f'<{len(values)}i').pack(*values))

    def write_u32(self, *values):
        self.stream.write(Struct(f'<{len(values)}I').pack(*values))

    def write_i64(self, *values):
        self.stream.write(Struct(f'<{len(values)}q').pack(*values))

    def write_u64(self, *values):
        self.stream.write(Struct(f'<{len(values)}Q').pack(*values))

    def write_f32(self, *values):
        self.stream.write(Struct(f'<{len(values)}f').pack(*values))
    
    def write_ul(self, *values): # For rst
        self.stream.write(Struct(f'<{len(values)}L').pack(*values))

    def write_ull(self, *values): # For rst
        self.stream.write(Struct(f'<{len(values)}Q').pack(*values))
    
    def write_vec2(self, *values):
        floats = [f for vec in values for f in vec]
        self.stream.write(Struct(f'<{len(floats)}f').pack(*floats))

    def write_vec3(self, *values):
        floats = [f for vec in values for f in vec]
        self.stream.write(Struct(f'<{len(floats)}f').pack(*floats))

    def write_vec4(self, *values):
        floats = [f for vec in values for f in vec]
        self.stream.write(Struct(f'<{len(floats)}f').pack(*floats))

    def write_quat(self, *values):
        floats = [f for quat in values for f in quat]
        self.stream.write(Struct(f'<{len(floats)}f').pack(*floats))

    def write_mtx4(self, mtx4):
        floats = [f for f in mtx4]
        self.stream.write(Struct('16f').pack(*floats))
    
    def write_s(self, value, encoding='ascii'):
        self.stream.write(value.encode(encoding))

    def write_s_padded(self, value, length, encoding='ascii'):
        if len(value) > length:
            value = value[:length]
        
        v = value.encode(encoding)
        self.stream.write(v + b'\x00'*(length-len(v)))

    def write_s_sized16(self, value, encoding='ascii'):
        v = value.encode(encoding)
        self.stream.write(Struct('H').pack(len(v)))
        self.stream.write(v)

    def write_s_sized32(self, value, encoding='ascii'):
        v = value.encode(encoding)
        self.stream.write(Struct('I').pack(len(v)))
        self.stream.write(v)

    def write_c_sep_0(self, value):
        s = b''
        for c in value.encode('ascii'):
            s += bytes([c])
            s += b'\x00'
        self.stream.write(s)



from io import BytesIO
from .stream import BinStream
from .structs import Vector
from enum import IntEnum, IntFlag
from random import uniform
from math import sqrt

class MAPGEOPlanarReflector:
    __slots__ = (
        'transform', 'plane', 'normal'
    )

    def __init__(self, transform=None, plane=None, normal=None):
        self.transform = transform
        self.plane = plane
        self.normal = normal

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

class MAPGEOBUcketGridFlag(IntFlag):
    HasFaceVisibilityFlags = 1 << 0

    def __json__(self):
        return self.__str__()

class MAPGEOBucket:
    __slots__ = (
        'max_stickout_x', 'max_stickout_z',
        'start_index', 'base_vertex',
        'inside_face_count', 'sticking_out_face_count'
    )

    def __init__(self, max_stickout_x=None, max_stickout_z=None, start_index=None, base_vertex=None, inside_face_count=None, sticking_out_face_count=None):
        self.max_stickout_x = max_stickout_x
        self.max_stickout_z = max_stickout_z
        self.start_index = start_index
        self.base_vertex = base_vertex
        self.inside_face_count = inside_face_count
        self.sticking_out_face_count = sticking_out_face_count
    
    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

class MAPGEOBucketGrid:
    __slots__ = (
        'hash',
        'min_x', 'min_z', 'max_x', 'max_z', 
        'max_stickout_x', 'max_stickout_z', 'bucket_size_x', 'bucket_size_z',
        'is_disabled', 'bucket_grid_flags',
        'buckets', 'vertices', 'indices',
        'face_layers'
    )

    def __init__(self, hash=None, min_x=None, min_z=None, max_x=None, max_z=None, max_stickout_x=None, max_stickout_z=None, bucket_size_x=None, bucket_size_z=None, is_disabled=None, bucket_grid_flags=None, buckets=None, vertices=None, indices=None, face_layers=None):
        self.hash = hash
        self.min_x = min_x
        self.min_z = min_z
        self.max_x = max_x
        self.max_z = max_z
        self.max_stickout_x = max_stickout_x
        self.max_stickout_z = max_stickout_z
        self.bucket_size_x = bucket_size_x
        self.bucket_size_z = bucket_size_z
        self.is_disabled = is_disabled
        self.bucket_grid_flags = bucket_grid_flags
        self.buckets = buckets
        self.vertices = vertices
        self.indices = indices
        self.face_layers = face_layers
    
    def __json__(self):
        dic = {key: getattr(self, key) for key in self.__slots__}
        dic['vertices'] = ('write only first vertex to save memory', dic['vertices'][0])
        dic['indices'] = ('not write to save memory')
        dic['face_layers'] = ('not write to save memory')
        return dic

class MAPGEOTextureOverride:
    __slots__ = ('index', 'path')

    def __init__(self, index=None, path=None):
        self.index = index
        self.path = path

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

class MAPGEOChannel:
    __slots__ = ('path', 'scale', 'offset')

    def __init__(self, path=None, scale=None, offset=None):
        self.path = path
        self.scale = scale
        self.offset = offset

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}
    
class MAPGEORender(IntFlag):
    IsDecal = 1 << 0
    HasEnvironmentDistortion = 1 << 1
    RenderOnlyIfEyeCandyOn = 1 << 2
    RenderOnlyIfEyeCandyOff = 1 << 3
    CreateShadowBuffer = 1 << 4
    CreateShadowMapMaterial = 1 << 5
    UnkCreateDepthBuffer2 = 1 << 6
    CreateDepthBuffer = 1 << 7

    def __json__(self):
        return self.__str__()

class MAPGEOQuality(IntFlag):
    VeryLow = 1 << 0
    Low = 1 << 1
    Medium = 1 << 2
    High = 1 << 3
    VeryHigh = 1 << 4

    def __json__(self):
        return self.__str__()

class MAPGEOLayer(IntFlag):
    Layer1 = 1 << 0
    Layer2 = 1 << 1
    Layer3 = 1 << 2
    Layer4 = 1 << 3
    Layer5 = 1 << 4
    Layer6 = 1 << 5
    Layer7 = 1 << 6
    Layer8 = 1 << 7

    def __json__(self):
        return self.__str__()
    

class MAPGEOSubmesh:
    __slots__ = (
        'name', 'hash',
        'index_start', 'index_count',
        'min_vertex', 'max_vertex'
    )

    def __init__(self, name=None, hash=None, index_start=None, index_count=None, min_vertex=None, max_vertex=None):
        self.name = name
        self.hash = hash
        self.index_start = index_start
        self.index_count = index_count
        self.min_vertex = min_vertex
        self.max_vertex = max_vertex
    
    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

class MAPGEOVertex:
    __slots__ = ('value')

    def __init__(self, value=None):
        self.value = value

    def __json__(self):
        return self.value

class MAPGEOModel:
    __slots__ = (
        'name',
        'vertex_buffer_count', 'vertex_description_id', 'vertex_buffer_ids', 'vertex_count', 'vertices', 
        'index_buffer_id', 'index_count', 'indices',
        'layer', 'quality', 'disable_backface_culling', 'is_bush', 'render', 'point_light', 'light_probe', 
        'bucket_grid_hash', 'submeshes',
        'bounding_box', 'matrix', 
        'baked_light', 'stationary_light', 'texture_overrides', 'texture_overrides_scale_offset'
    )

    def __init__(self, name=None, vertex_buffer_count=None, vertex_description_id=None, vertex_buffer_ids=None, vertex_count=None, vertices=None, index_buffer_id=None, index_count=None, indices=None, layer=None, quality=None, disable_backface_culling=None, is_bush=None, render=None, point_light=None, light_probe=None, bucket_grid_hash=None, submeshes=None, bounding_box=None, matrix=None, baked_light=None, stationary_light=None, texture_overrides=None, texture_overrides_scale_offset=None):
        self.name = name
        self.vertex_buffer_count = vertex_buffer_count
        self.vertex_description_id = vertex_description_id
        self.vertex_buffer_ids = vertex_buffer_ids
        self.vertex_count = vertex_count
        self.vertices = vertices
        self.index_buffer_id = index_buffer_id
        self.index_count = index_count
        self.indices = indices
        self.layer = layer
        self.quality = quality
        self.disable_backface_culling = disable_backface_culling
        self.is_bush = is_bush
        self.render = render
        self.point_light = point_light
        self.light_probe = light_probe
        self.bucket_grid_hash = bucket_grid_hash
        self.submeshes = submeshes
        self.bounding_box = bounding_box
        self.matrix = matrix
        self.baked_light = baked_light
        self.stationary_light = stationary_light
        self.texture_overrides = texture_overrides
        self.texture_overrides_scale_offset = texture_overrides_scale_offset

    def __json__(self):
        dic = {key: getattr(self, key) for key in self.__slots__}
        dic['vertices'] = ('write only first vertex to save memory', dic['vertices'][0])
        dic['indices'] = ('not write to save memory')
        return dic
    
class MAPGEOVertexElementName(IntEnum):
    Position = 0
    BlendWeight = 1
    Normal = 2
    FogCoordinate = 3
    PrimaryColor = 4
    SecondaryColor = 5
    BlendIndex = 6
    Texcoord0 = 7
    Texcoord1 = 8
    Texcoord2 = 9
    Texcoord3 = 10
    Texcoord4 = 11
    Texcoord5 = 12
    Texcoord6 = 13
    Texcoord7 = 14
    Tangent = 15

    def __json__(self):
        return self.name

class MAPGEOVertexElementFormat(IntEnum):
    X_Float32 = 0
    XY_Float32 = 1
    XYZ_Float32 = 2
    XYZW_Float32 = 3
    BGRA_Packed8888 = 4
    ZYXW_Packed8888 = 5
    RGBA_Packed8888 = 6
    XYZW_Packed8888 = 7
    XY_Float32_2 = 8

    def __json__(self):
        return self.name
   
class MAPGEOVertexElement:
    __slots__ = ('name', 'format')

    def __init__(self, name=None, format=None):
        self.name = name
        self.format = format

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

class MAPGEOVertexUsage(IntEnum):
    Static = 0
    Dynamic = 1
    Stream = 2

    def __json__(self):
        return self.name

class MAPGEOVertexDescription:
    __slots__ = ( 'usage', 'elements')

    def __init__(self, usage=None, elements=None):
        self.usage = usage
        self.elements = elements

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

class MAPGEOHelper:
    MGVertexFormatToPyValues = {
        # mapgeo vertex format: (python struct format, bytes size, unpacked items size, unpacked type)
        MAPGEOVertexElementFormat.X_Float32: ('f', 4, 1, float),  
        MAPGEOVertexElementFormat.XY_Float32: ('2f', 8, 2, Vector),  
        MAPGEOVertexElementFormat.XYZ_Float32: ('3f', 12, 3, Vector), 
        MAPGEOVertexElementFormat.XYZW_Float32: ('4f', 16, 4, Vector),  
        MAPGEOVertexElementFormat.BGRA_Packed8888: ('4B', 4, 4, tuple), 
        MAPGEOVertexElementFormat.ZYXW_Packed8888: ('4B', 4, 4, tuple), 
        MAPGEOVertexElementFormat.RGBA_Packed8888: ('4B', 4, 4, tuple), 
        MAPGEOVertexElementFormat.XYZW_Packed8888: ('4B', 4, 4, tuple),  
        MAPGEOVertexElementFormat.XY_Float32_2: ('2f', 8, 2, Vector)  
    }

class MAPGEO:
    __slots__ = (
        'signature', 'version', 'texture_overrides', 'vertex_descriptions',
        'models', 'bucket_grids', 'planar_reflectors'
    )
    
    def __init__(self, signature=None, version=None, texture_overrides=None, vertex_descriptions=None, models=None, bucket_grids=None, planar_reflectors=None):
        self.signature = signature
        self.version = version
        self.texture_overrides = texture_overrides
        self.vertex_descriptions = vertex_descriptions
        self.models = models
        self.bucket_grids = bucket_grids
        self.planar_reflectors = planar_reflectors

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

    def stream(self, path, mode, raw=None):
        if raw != None:
            if raw == True:  # the bool True value
                return BinStream(BytesIO())
            else:
                return BinStream(BytesIO(raw))
        return BinStream(open(path, mode))
    
    def read(self, path, raw=None):
        with self.stream(path, 'rb', raw) as bs:
            self.signature, = bs.read_s(4)
            if self.signature != 'OEGM':
                raise Exception(
                    f'pyRitoFile: Failed: Read MAPGEO {path}: Wrong signature file: {self.signature}')
            self.version, = bs.read_u32()
            if self.version not in (5, 6, 7, 9, 11, 12, 13, 14, 15, 17):
                raise Exception(
                    f'pyRitoFile: Failed: Read MAPGEO {path}: Unsupported file version: {self.version}')

            # point light
            use_seperate_point_lights = False
            if self.version < 7:
                use_seperate_point_lights, = bs.read_b()

            # texture overrides
            if self.version >= 17:
                texture_override_count, = bs.read_u32()
                self.texture_overrides = [MAPGEOTextureOverride() for i in range(texture_override_count)]
                for texture_override in self.texture_overrides:
                    texture_override.index, = bs.read_u32()
                    texture_override.path, = bs.read_s_sized32()
            else:
                self.texture_overrides = []
                if self.version >= 9:
                    texture_override = MAPGEOTextureOverride()
                    texture_override.index = 0
                    texture_override.path, = bs.read_s_sized32()
                    self.texture_overrides.append(texture_override)
                    if self.version >= 11:
                        texture_override = MAPGEOTextureOverride()
                        texture_override.index = 1
                        texture_override.path, = bs.read_s_sized32()
                        self.texture_overrides.append(texture_override)

            # vertex descriptions
            vertex_description_count, = bs.read_u32()
            self.vertex_descriptions = [MAPGEOVertexDescription() for i in range(vertex_description_count)]
            for vertex_description in self.vertex_descriptions:
                vertex_description.usage = MAPGEOVertexUsage(bs.read_u32()[0])
                element_count, = bs.read_u32()
                unpacked_u32s = bs.read_u32(element_count * 2)
                vertex_description.elements = [MAPGEOVertexElement() for i in range(element_count)]
                for element_id, element in enumerate(vertex_description.elements):
                    element.name = MAPGEOVertexElementName(unpacked_u32s[element_id*2])
                    element.format = MAPGEOVertexElementFormat(unpacked_u32s[element_id*2+1])
                bs.pad(8 * (15 - element_count))  # pad empty vertex descriptions

            # vertex buffers 
            # -> can not read now because need vertex descriptions 
            # -> read vertex buffer's offset instead
            vertex_buffer_count, = bs.read_u32()
            vertex_buffer_offsets = [None]*vertex_buffer_count
            for i in range(vertex_buffer_count):
                if self.version >= 13:
                    bs.pad(1)  # layer 
                vertex_buffer_size, = bs.read_u32() # size = byte count
                vertex_buffer_offsets[i] = bs.tell()
                bs.pad(vertex_buffer_size)
        
            # index buffers
            # -> can straight read because its just int array
            index_buffer_count, = bs.read_u32()
            index_buffers = [None]*index_buffer_count
            for i in range(index_buffer_count):
                if self.version >= 13:
                    bs.pad(1)  # layer
                index_buffer_size, = bs.read_u32() # size = byte count
                index_buffers[i] = bs.read_u16(index_buffer_size // 2)

            # read models
            unpacked_vertex_buffers = [None]*vertex_buffer_count # for skip reading same vertex buffer
            model_count, = bs.read_u32()
            self.models = [MAPGEOModel() for i in range(model_count)]
            for model_id, model in enumerate(self.models):
                print(f'read {model_id}')
                # model name
                if self.version < 12:
                    model.name, = bs.read_s_sized32()
                else:
                    model.name = f'MapGeo_Instance_{model_id}'

                # model vertices
                model.vertex_count, model.vertex_buffer_count, model.vertex_description_id = bs.read_u32(3)
                # read vertex buffer into unpacked vbs
                model.vertex_buffer_ids = bs.read_i32(model.vertex_buffer_count)
                for i, vertex_buffer_id in enumerate(model.vertex_buffer_ids):
                    vertex_description = self.vertex_descriptions[model.vertex_description_id+i]
                    # skip unpacked vertex buffers
                    if unpacked_vertex_buffers[vertex_buffer_id] != None:
                        continue
                    unpacked_vertex_buffers[vertex_buffer_id] = []
                    # vertex python format & size through vertex descriptions of single vertex (yes only 1)
                    vertex_python_format = ''
                    vertex_byte_size = 0
                    for element in vertex_description.elements:
                        python_format, byte_size, _, _ = MAPGEOHelper.MGVertexFormatToPyValues[element.format]
                        vertex_python_format += python_format
                        vertex_byte_size += byte_size
                    # save the current offset (we are in middle of model reading)
                    # jump to vertex buffers place through saved vertex buffer offset
                    return_offset = bs.tell()
                    bs.seek(vertex_buffer_offsets[vertex_buffer_id])
                    # read whole vertex buffers with vertex format & size
                    # every vertex use same format so just multiply with vertex count
                    unpacked_vertex_buffers[vertex_buffer_id] = bs.read_fmt(
                        fmt = vertex_python_format*model.vertex_count,
                        fmt_size = vertex_byte_size*model.vertex_count
                    )
                    # return to model reading after read vertices
                    bs.seek(return_offset)
                # now with unpacked vbs, set values for vertex
                model.vertices = [MAPGEOVertex() for i in range(model.vertex_count)]
                for i, vertex_buffer_id in enumerate(model.vertex_buffer_ids):
                    vertex_description = self.vertex_descriptions[model.vertex_description_id+i]
                    unpacked_vb = unpacked_vertex_buffers[vertex_buffer_id]
                    current_index = 0
                    for vertex in model.vertices:
                        vertex.value = {}
                        for element in vertex_description.elements:
                            _, _, unpacked_item_size, unpacked_type = MAPGEOHelper.MGVertexFormatToPyValues[element.format]
                            unpacked_item_value = unpacked_vb[current_index:current_index+unpacked_item_size]
                            vertex.value[element.name.name] = unpacked_type(*unpacked_item_value) if unpacked_type is not tuple else unpacked_type(unpacked_item_value)
                            current_index += unpacked_item_size

                # model indices
                model.index_count, model.index_buffer_id = bs.read_u32(2)
                model.indices = index_buffers[model.index_buffer_id]

                # layer - above or version 13
                if self.version >= 13:
                    model.layer = MAPGEOLayer(bs.read_u8()[0])
                
                # bucket grid hash
                if self.version >= 15:
                    model.bucket_grid_hash, = bs.read_u32()

                # submeshes
                submesh_count, = bs.read_u32()
                model.submeshes = [MAPGEOSubmesh() for i in range(submesh_count)]
                for submesh in model.submeshes:
                    submesh.hash, = bs.read_u32()
                    submesh.name, = bs.read_s_sized32()
                    submesh.index_start, submesh.index_count, submesh.min_vertex, submesh.max_vertex = bs.read_u32(
                        4)

                if self.version != 5:
                    model.disable_backface_culling, = bs.read_b()

                # bounding box
                model.bounding_box = (bs.read_vec3()[0], bs.read_vec3()[0])

                # transform matrix
                model.matrix, = bs.read_mtx4()

                # quality 
                model.quality = MAPGEOQuality(bs.read_u8()[0])

                # layer - below version 13
                if self.version >= 7 and self.version <= 12:
                    model.layer = MAPGEOLayer(bs.read_u8()[0])

                if self.version >= 11:
                    if self.version < 14:
                        model.render = MAPGEORender(bs.read_u8()[0])
                    else:
                        model.is_bush, = bs.read_b()
                        if self.version < 16:
                            model.render = MAPGEORender(bs.read_u8()[0])
                        else:
                            model.render = MAPGEORender(bs.read_u16()[0])

                # point light
                if self.version < 7 and use_seperate_point_lights:
                    model.point_light, = bs.read_vec3()

                if self.version < 9:
                    # light probes
                    # 27 floats, 9 for each RGB channel
                    model.light_probe = (bs.read_f32(9), bs.read_f32(9), bs.read_f32(9))

                # channels
                model.baked_light = MAPGEOChannel()
                model.baked_light.path, = bs.read_s_sized32()
                model.baked_light.scale = bs.read_f32(2)
                model.baked_light.offset = bs.read_f32(2)

                if self.version >= 9:
                    model.stationary_light = MAPGEOChannel()
                    model.stationary_light.path, = bs.read_s_sized32()
                    model.stationary_light.scale = bs.read_f32(2)
                    model.stationary_light.offset = bs.read_f32(2)

                    # texture override
                    if self.version >= 12:
                        if self.version < 17:
                            model.texture_overrides = [MAPGEOTextureOverride()]
                            model.texture_overrides[0].index = 0
                            model.texture_overrides[0].path, = bs.read_s_sized32()
                            model.texture_overrides_scale_offset = bs.read_f32(4)
                        else:
                            texture_override_count, = bs.read_u32()
                            model.texture_overrides = [MAPGEOTextureOverride() for i in range(texture_override_count)]
                            for texture_override in model.texture_overrides:
                                texture_override.index, = bs.read_u32()
                                texture_override.path, = bs.read_s_sized32()
                            model.texture_overrides_scale_offset = bs.read_f32(4)

            # for modded file with no bucket grid, planar reflector: stop reading
            # (exported by lemon3d/lol_maya)
            current = bs.tell()
            end = bs.end()
            if current == end:
                return

            # bucket grids
            # version 15: multi bucket grids, below version 15 only store one
            bucket_grid_count = bs.read_u32()[0] if self.version >= 15 else 1
            self.bucket_grids = [MAPGEOBucketGrid() for i in range(bucket_grid_count)]
            for bucket_grid in self.bucket_grids:
                # hash - version 15
                if self.version >= 15:
                    bucket_grid.hash, = bs.read_u32()
                bucket_grid.min_x, bucket_grid.min_z, bucket_grid.max_x, bucket_grid.max_z, bucket_grid.max_stickout_x, bucket_grid.max_stickout_z, bucket_grid.bucket_size_x, bucket_grid.bucket_size_z = bs.read_f32(8)
                bucket_count, = bs.read_u16()
                bucket_grid.is_disabled, = bs.read_b()
                bucket_grid.bucket_grid_flags = MAPGEOBUcketGridFlag(bs.read_u8()[0])
                vertex_count, index_count = bs.read_u32(2)
                if not bucket_grid.is_disabled:
                    bucket_grid.vertices = bs.read_vec3(vertex_count)
                    bucket_grid.indices = bs.read_u16(index_count)
                    bucket_grid.buckets = [[MAPGEOBucket() for j in range(bucket_count)] for i in range(bucket_count)]
                    for bucket_row in bucket_grid.buckets:
                        for bucket in bucket_row:
                            bucket.max_stickout_x, bucket.max_stickout_z = bs.read_f32(2)
                            bucket.start_index, bucket.base_vertex = bs.read_u32(2)
                            bucket.inside_face_count, bucket.sticking_out_face_count = bs.read_u16(2)
            
                    if MAPGEOBUcketGridFlag.HasFaceVisibilityFlags in bucket_grid.bucket_grid_flags:
                        unpacked_u8s = bs.read_u8(index_count // 3)
                        bucket_grid.face_layers = [MAPGEOLayer(unpacked_u8) for unpacked_u8 in unpacked_u8s]
            
            if self.version >= 13:
                pr_count, = bs.read_u32()
                self.planar_reflectors = [MAPGEOPlanarReflector() for i in range(pr_count)]
                for planar_reflector in self.planar_reflectors:
                    planar_reflector.transform = bs.read_mtx4,
                    planar_reflector.plane = bs.read_vec3(2)
                    planar_reflector.normal, = bs.read_vec3()


    def write(self, path, version, raw=None):
        if version not in (13, 17):
            raise Exception(
                f'pyRitoFile: Failed: Write MAPGEO {path}: Unsupported file version: {version}')
        self.version = version

        with self.stream(path, 'wb', raw) as bs:
            # prepare stuffs
            self.vertex_descriptions = []
            vertex_buffers = []
            index_buffers = []
            for model in self.models:
                # vertex descriptions
                vertex_formats = ''
                vertex_size = 0
                vertex_description = MAPGEOVertexDescription()
                vertex_description.usage = MAPGEOVertexUsage.Static
                vertex_description.elements = []
                vertex = model.vertices[0]
                if MAPGEOVertexElementName.Position.name in vertex.value:
                    element = MAPGEOVertexElement()
                    element.name = MAPGEOVertexElementName.Position
                    element.format = MAPGEOVertexElementFormat.XYZ_Float32
                    vertex_description.elements.append(element)
                    vertex_formats += MAPGEOHelper.MGVertexFormatToPyValues[MAPGEOVertexElementFormat.XYZ_Float32][0]
                    vertex_size += MAPGEOHelper.MGVertexFormatToPyValues[MAPGEOVertexElementFormat.XYZ_Float32][1]
                if MAPGEOVertexElementName.Normal.name in vertex.value:
                    element = MAPGEOVertexElement()
                    element.name = MAPGEOVertexElementName.Normal
                    element.format = MAPGEOVertexElementFormat.XYZ_Float32
                    vertex_description.elements.append(element)
                    vertex_formats += MAPGEOHelper.MGVertexFormatToPyValues[MAPGEOVertexElementFormat.XYZ_Float32][0]
                    vertex_size += MAPGEOHelper.MGVertexFormatToPyValues[MAPGEOVertexElementFormat.XYZ_Float32][1]
                if MAPGEOVertexElementName.PrimaryColor.name in vertex.value:
                    element = MAPGEOVertexElement()
                    element.name = MAPGEOVertexElementName.PrimaryColor
                    element.format = MAPGEOVertexElementFormat.BGRA_Packed8888
                    vertex_description.elements.append(element)
                    vertex_formats += MAPGEOHelper.MGVertexFormatToPyValues[MAPGEOVertexElementFormat.BGRA_Packed8888][0]
                    vertex_size += MAPGEOHelper.MGVertexFormatToPyValues[MAPGEOVertexElementFormat.BGRA_Packed8888][1]
                if MAPGEOVertexElementName.Texcoord0.name in vertex.value:
                    element = MAPGEOVertexElement()
                    element.name = MAPGEOVertexElementName.Texcoord0
                    element.format = MAPGEOVertexElementFormat.XY_Float32
                    vertex_description.elements.append(element)
                    vertex_formats += MAPGEOHelper.MGVertexFormatToPyValues[MAPGEOVertexElementFormat.XY_Float32][0]
                    vertex_size += MAPGEOHelper.MGVertexFormatToPyValues[MAPGEOVertexElementFormat.XY_Float32][1]
                if self.version > 13 and model.is_bush:
                    element = MAPGEOVertexElement()
                    element.name = MAPGEOVertexElementName.Texcoord5
                    element.format = MAPGEOVertexElementFormat.XYZ_Float32
                    vertex_description.elements.append(element)
                    vertex_formats += MAPGEOHelper.MGVertexFormatToPyValues[MAPGEOVertexElementFormat.XYZ_Float32][0]
                    vertex_size += MAPGEOHelper.MGVertexFormatToPyValues[MAPGEOVertexElementFormat.XYZ_Float32][1]
                if MAPGEOVertexElementName.Texcoord7.name in vertex.value:
                    element = MAPGEOVertexElement()
                    element.name = MAPGEOVertexElementName.Texcoord7
                    element.format = MAPGEOVertexElementFormat.XY_Float32
                    vertex_description.elements.append(element)
                    vertex_formats += MAPGEOHelper.MGVertexFormatToPyValues[MAPGEOVertexElementFormat.XY_Float32][0]
                    vertex_size += MAPGEOHelper.MGVertexFormatToPyValues[MAPGEOVertexElementFormat.XY_Float32][1]
                self.vertex_descriptions.append(vertex_description)
                vertex_formats = vertex_formats * len(model.vertices)
                vertex_size = vertex_size * len(model.vertices)
                # vertex values
                vertex_values = []
                boudingbox_min = Vector(float("inf"), float("inf"), float("inf"))
                boudingbox_max = Vector(float("-inf"), float("-inf"), float("-inf"))
                for vertex in model.vertices:
                    for element in vertex_description.elements:
                        vertex_value = vertex.value[element.name]
                        vertex_values.extend(
                            value for value in vertex_value
                        )
                        if element.name == MAPGEOVertexElementName.Position.name:
                            # find bounding box
                            position = vertex_value
                            if boudingbox_min.x > position.x:
                                boudingbox_min.x = position.x
                            if boudingbox_min.y > position.y:
                                boudingbox_min.y = position.y
                            if boudingbox_min.z > position.z:
                                boudingbox_min.z = position.z
                            if boudingbox_max.x < position.x:
                                boudingbox_max.x = position.x
                            if boudingbox_max.y < position.y:
                                boudingbox_max.y = position.y
                            if boudingbox_max.z < position.z:
                                boudingbox_max.z = position.z
                                
                # vertex buffers
                vertex_buffers.append((
                    model.layer,
                    vertex_size,
                    vertex_formats,
                    vertex_values
                ))
                # index buffers
                index_buffers.append((
                    model.layer,
                    model.indices
                ))
                # bounding box
                model.bounding_box = (boudingbox_min, boudingbox_min)

            # start to write mapgeo
            # signature, version
            bs.write_s('OEGM')
            bs.write_u32(self.version)
            # texture overrides
            if self.version > 13:
                bs.write_u32(0) # texture override count
            else:
                bs.write_u32(0, 0) # length of texture override path 1 & 2
            
            # vertex descriptions
            bs.write_u32(len(self.vertex_descriptions))
            for vertex_description in self.vertex_descriptions:
                bs.write_u32(vertex_description.usage.value) 
                element_count = len(vertex_description.elements)
                bs.write_u32(element_count) 
                for element in vertex_description.elements:
                    bs.write_u32(element.name.value, element.format.value)
                # fill remaining empty vertex descriptions
                fill = (MAPGEOVertexElementName.Position.value, MAPGEOVertexElementFormat.XYZ_Float32.value) * ( 15 - element_count)
                bs.write_u32(*fill)

            # vertex buffers
            bs.write_u32(len(vertex_buffers))
            for layer, vertex_size, vertex_formats, vertex_values in vertex_buffers:
                bs.write_u8(layer.value)
                bs.write_u32(vertex_size)
                bs.write_fmt(vertex_formats, *vertex_values)

            # index buffers
            bs.write_u32(len(index_buffers))
            for layer, indices in index_buffers:
                bs.write_u8(layer.value)
                bs.write_u32(len(indices) * 2)
                bs.write_u16(*indices)

            # model
            bs.write_u32(len(self.models))
            for model_id, model in enumerate(self.models):
                print(f'write {model_id}')
                bs.write_u32(
                    len(model.vertices),  # vertex count
                    1,  # vertex buffer count
                    model_id,  # vertex description id
                    model_id,  # vertex buffer id
                    len(model.indices),  # index count
                    model_id  # index buffer id
                )

                 # layer
                bs.write_u8(model.layer.value)

                # bucket grid hash
                if self.version > 13:
                    bs.write_u32(model.bucket_grid_hash)
                
                # submeshes
                bs.write_u32(len(model.submeshes))
                for submesh in model.submeshes:
                    bs.write_u32(0)  # hash, always 0?
                    bs.write_s_sized32(submesh.name)
                    bs.write_u32(
                        submesh.index_start, submesh.index_count, submesh.min_vertex, submesh.max_vertex)

                # flip normals
                bs.write_u8(0)

                # bounding box
                bs.write_vec3(*model.bounding_box)

                # transform
                bs.write_f32(*model.matrix)

                # quality, 31 = all quality
                bs.write_u8(31)

                # bush, render flag
                if self.version > 13:
                    bs.write_u8(model.is_bush) # bush
                    bs.write_u16(0) # render flag
                else:
                    bs.write_u8(0) # render flag

                # baked light
                if model.baked_light not in (None, ''):
                    bs.write_s_sized32(model.baked_light.path)
                    bs.write_f32(1.0, 1.0, 0.0, 0.0)
                else:
                    bs.write_u32(0)
                    bs.write_f32(0.0, 0.0, 0.0, 0.0)
                # stationary light
                bs.write_u32(0)
                bs.write_f32(0.0, 0.0, 0.0, 0.0)
                # texture override
                if self.version > 13:
                    bs.write_u32(len(model.texture_overrides))
                    for texture_override in model.texture_overrides:
                        bs.write_u32(texture_override.index)
                        bs.write_s_sized32(texture_override.path)
                    bs.write_f32(*model.texture_overrides_scale_offset)
                else:
                    bs.write_u32(0)
                    bs.write_f32(0.0, 0.0, 0.0, 0.0)

            # bucket grid
            bucket_grid_count = len(self.bucket_grids)
            if bucket_grid_count > 0:
                if self.version > 13:
                    bs.write_u32(bucket_grid_count)
                for bucket_grid in self.bucket_grids:
                    if self.version > 13:
                        bs.write_u32(bucket_grid.hash)
                    bs.write_f32(
                        bucket_grid.min_x, 
                        bucket_grid.min_z, 
                        bucket_grid.max_x, 
                        bucket_grid.max_z, 
                        bucket_grid.max_stickout_x, 
                        bucket_grid.max_stickout_z, 
                        bucket_grid.bucket_size_x, 
                        bucket_grid.bucket_size_z
                    ) 
                    bs.write_u16(int(sqrt(len(bucket_grid.buckets)//20))) # bucket size
                    bs.write_b(bucket_grid.is_disabled)
                    bs.write_u8(bucket_grid.bucket_grid_flags.value)
                    bs.write_u32(len(bucket_grid.vertices)//12)
                    bs.write_u32(len(bucket_grid.indices)//2)
                    if not bucket_grid.is_disabled:
                        bs.write_vec3(*bucket_grid.vertices)
                        bs.write_u16(*bucket_grid.indices)
                        for bucket_row in bucket_grid.buckets:
                            for bucket in bucket_row:
                                bs.write_f32(bucket.max_stickout_x, bucket.max_stickout_z)
                                bs.write_u32(bucket.start_index, bucket.base_vertex)
                                bs.write_u16(bucket.inside_face_count, bucket.sticking_out_face_count)
                        if MAPGEOBUcketGridFlag.HasFaceVisibilityFlags in bucket_grid.bucket_grid_flags:
                            bs.write_u8(*[face_layer.value for face_layer in bucket_grid.face_layers])

            if self.planar_reflectors != None:
                bs.write_u32(len(self.planar_reflectors))
                for planar_reflector in self.planar_reflectors:
                    bs.write_mtx4(planar_reflector.transform)
                    bs.read_vec3(*planar_reflector.plane)
                    bs.write_vec3(planar_reflector.normal)

            return bs.raw() if raw else None


from math import sqrt, acos, sin


class Vector:
    __slots__ = ('x', 'y', 'z', 'w')

    def __init__(self, x, y, z=None, w=None):
        self.x = x
        self.y = y
        self.z = z
        self.w = w

    def __iter__(self):
        yield self.x
        yield self.y
        if self.z != None:
            yield self.z
        if self.w != None:
            yield self.w

    def __str__(self):
        s = f'{self.x:.4f} {self.y:.4f}'
        if self.z != None:
            s += f' {self.z:.4f}'
        if self.w != None:
            s += f' {self.w:.4f}'
        return s

    def __json__(self):
        return [v for v in self]
    
    def lerp(vec1, vec2, weight):  
        if weight < 0.005:
            return vec1
        if weight > 0.995:
            return vec2      
        return Vector(
            vec1.x + (vec2.x - vec1.x) * weight,
            vec1.y + (vec2.y - vec1.y) * weight,
            vec1.z + (vec2.z - vec1.z) * weight
        )


class Quaternion:
    __slots__ = ('x', 'y', 'z', 'w')

    def __init__(self, x, y, z, w):
        self.x = x
        self.y = y
        self.z = z
        self.w = w

    def __iter__(self):
        yield self.x
        yield self.y
        yield self.z
        yield self.w


    def __mul__(self, other):
        if type(other) in (float, int):
            self.x *= other
            self.y *= other
            self.z *= other
            self.w *= other
        return self

    def __str__(self):
        return f'{self.x:.4f} {self.y:.4f} {self.z:.4f} {self.w:.4f}'

    def __json__(self):
        return [v for v in self]
    
    def slerp(quat1, quat2, weight):
        if weight < 0.005:
            return quat1
        if weight > 0.995:
            return quat2    
        epsilon = 1e-6
        cos_omega = quat1.x * quat2.x + quat1.y * quat2.y + quat1.z * quat2.z + quat1.w * quat2.w
        flip = False
        if cos_omega < 0.0:
            flip = True
            cos_omega = -cos_omega
        if cos_omega > (1.0 - epsilon):
            s1 = 1.0 - weight
            s2 = -weight if flip else weight
        else:
            omega = acos(cos_omega)
            inv_sin_omega = 1 / sin(omega)
            s1 = sin((1.0 - weight) * omega) * inv_sin_omega
            s2 = -sin(weight * omega) * inv_sin_omega if flip else sin(weight * omega) * inv_sin_omega
        return Quaternion(
            s1 * quat1.x + s2 * quat2.x,
            s1 * quat1.y + s2 * quat2.y,
            s1 * quat1.z + s2 * quat2.z,
            s1 * quat1.w + s2 * quat2.w,
        )        


class Matrix4:
    __slots__ = (
        'a', 'b', 'c', 'd',
        'e', 'f', 'g', 'h',
        'i', 'j', 'k', 'l',
        'm', 'n', 'o', 'p'
    )

    def __init__(self, *values):
        if len(values) == 16:
            (self.a, self.b, self.c, self.d,
             self.e, self.f, self.g, self.h,
             self.i, self.j, self.k, self.l,
             self.m, self.n, self.o, self.p) = values
        else:
            self.a = self.f = self.k = self.p = 1.0
            self.b = self.c = self.d = self.e = self.g = self.h = self.i = self.j = self.l = self.m = self.n = self.o = 0.0

    def __getitem__(self, index):
        return (
            self.a, self.b, self.c, self.d,
            self.e, self.f, self.g, self.h,
            self.i, self.j, self.k, self.l,
            self.m, self.n, self.o, self.p
        )[index]

    def __setitem__(self, index, value):
        values = self[:]
        values[index] = value
        (
            self.a, self.b, self.c, self.d,
            self.e, self.f, self.g, self.h,
            self.i, self.j, self.k, self.l,
            self.m, self.n, self.o, self.p
        ) = values

    def __iter__(self):
        values = (
            self.a, self.b, self.c, self.d,
            self.e, self.f, self.g, self.h,
            self.i, self.j, self.k, self.l,
            self.m, self.n, self.o, self.p
        )
        for v in values:
            yield v

    def __str__(self):
        return (
            f'{self.a:.4f} {self.b:.4f} {self.c:.4f} {self.d:.4f}\n'
            f'{self.e:.4f} {self.f:.4f} {self.g:.4f} {self.h:.4f}\n'
            f'{self.i:.4f} {self.j:.4f} {self.k:.4f} {self.l:.4f}\n'
            f'{self.m:.4f} {self.n:.4f} {self.o:.4f} {self.p:.4f}'
        )

    def __json__(self):
        return [v for v in self]

    def __mul__(self, other):
        return Matrix4(
            self.a * other.a + self.b * other.e + self.c * other.i + self.d * other.m,
            self.a * other.b + self.b * other.f + self.c * other.j + self.d * other.n,
            self.a * other.c + self.b * other.g + self.c * other.k + self.d * other.o,
            self.a * other.d + self.b * other.h + self.c * other.l + self.d * other.p,

            self.e * other.a + self.f * other.e + self.g * other.i + self.h * other.m,
            self.e * other.b + self.f * other.f + self.g * other.j + self.h * other.n,
            self.e * other.c + self.f * other.g + self.g * other.k + self.h * other.o,
            self.e * other.d + self.f * other.h + self.g * other.l + self.h * other.p,

            self.i * other.a + self.j * other.e + self.k * other.i + self.l * other.m,
            self.i * other.b + self.j * other.f + self.k * other.j + self.l * other.n,
            self.i * other.c + self.j * other.g + self.k * other.k + self.l * other.o,
            self.i * other.d + self.j * other.h + self.k * other.l + self.l * other.p,

            self.m * other.a + self.n * other.e + self.o * other.i + self.p * other.m,
            self.m * other.b + self.n * other.f + self.o * other.j + self.p * other.n,
            self.m * other.c + self.n * other.g + self.o * other.k + self.p * other.o,
            self.m * other.d + self.n * other.h + self.o * other.l + self.p * other.p
        )

    def inverse(self):
        d = (
            (self.a * self.f - self.e * self.b)
            * (self.k * self.p - self.o * self.l)
            - (self.a * self.j - self.i * self.b)
            * (self.g * self.p - self.o * self.h)
            + (self.a * self.n - self.m * self.b)
            * (self.g * self.l - self.k * self.h)
            + (self.e * self.j - self.i * self.f)
            * (self.c * self.p - self.o * self.d)
            - (self.e * self.n - self.m * self.f)
            * (self.c * self.l - self.k * self.d)
            + (self.i * self.n - self.m * self.j)
            * (self.c * self.h - self.g * self.d)
        )

        inv = Matrix4()
        if abs(d) >= 0.001:
            d = 1.0 / d

            inv.a = d * (self.f * (self.k * self.p - self.o * self.l) + self.j * (
                self.o * self.h - self.g * self.p) + self.n * (self.g * self.l - self.k * self.h))
            inv.e = d * (self.g * (self.i * self.p - self.m * self.l) + self.k * (
                self.m * self.h - self.e * self.p) + self.o * (self.e * self.l - self.i * self.h))
            inv.i = d * (self.h * (self.i * self.n - self.m * self.j) + self.l * (
                self.m * self.f - self.e * self.n) + self.p * (self.e * self.j - self.i * self.f))
            inv.m = d * (self.e * (self.n * self.k - self.j * self.o) + self.i * (
                self.f * self.o - self.n * self.g) + self.m * (self.j * self.g - self.f * self.k))

            inv.b = d * (self.j * (self.c * self.p - self.o * self.d) + self.n * (
                self.k * self.d - self.c * self.l) + self.b * (self.o * self.l - self.k * self.p))
            inv.f = d * (self.k * (self.a * self.p - self.m * self.d) + self.o * (
                self.i * self.d - self.a * self.l) + self.c * (self.m * self.l - self.i * self.p))
            inv.j = d * (self.l * (self.a * self.n - self.m * self.b) + self.p * (
                self.i * self.b - self.a * self.j) + self.d * (self.m * self.j - self.i * self.n))
            inv.n = d * (self.i * (self.n * self.c - self.b * self.o) + self.m * (
                self.b * self.k - self.j * self.c) + self.a * (self.j * self.o - self.n * self.k))

            inv.c = d * (self.n * (self.c * self.h - self.g * self.d) + self.b * (
                self.g * self.p - self.o * self.h) + self.f * (self.o * self.d - self.c * self.p))
            inv.g = d * (self.o * (self.a * self.h - self.e * self.d) + self.c * (
                self.e * self.p - self.m * self.h) + self.g * (self.m * self.d - self.a * self.p))
            inv.k = d * (self.p * (self.a * self.f - self.e * self.b) + self.d * (
                self.e * self.n - self.m * self.f) + self.h * (self.m * self.b - self.a * self.n))
            inv.o = d * (self.m * (self.f * self.c - self.b * self.g) + self.a * (
                self.n * self.g - self.f * self.o) + self.e * (self.b * self.o - self.n * self.c))

            inv.d = d * (self.b * (self.k * self.h - self.g * self.l) + self.f * (
                self.c * self.l - self.k * self.d) + self.j * (self.g * self.d - self.c * self.h))
            inv.h = d * (self.c * (self.i * self.h - self.e * self.l) + self.g * (
                self.a * self.l - self.i * self.d) + self.k * (self.e * self.d - self.a * self.h))
            inv.l = d * (self.d * (self.i * self.f - self.e * self.j) + self.h * (
                self.a * self.j - self.i * self.b) + self.l * (self.e * self.b - self.a * self.f))
            inv.p = d * (self.a * (self.f * self.k - self.j * self.g) + self.e * (
                self.j * self.c - self.b * self.k) + self.i * (self.b * self.g - self.f * self.c))
        # else: return identity anyway
        return inv

    def decompose(self):
        # this only support scale (1.0, 1.0, 1.0)
        # but we only use for update old skl, so its enough
        translate = Vector(self.m, self.n, self.o)

        scale = Vector(
            self.p * sqrt(self.a**2 + self.b**2 + self.c**2),
            self.p * sqrt(self.e**2 + self.f**2 + self.g**2),
            self.p * sqrt(self.i**2 + self.j**2 + self.k**2)
        )

        r_mat = Matrix4(
            *(
                self.a/scale.x, self.b/scale.y, self.c/scale.z, 0,
                self.e/scale.x, self.f/scale.y, self.g/scale.z, 0,
                self.i/scale.x, self.j/scale.y, self.k/scale.z, 0,
                0, 0, 0, 1
            )
        )
        dott = (r_mat.b * r_mat.g - r_mat.c * r_mat.f) * r_mat.i \
            + (r_mat.c * r_mat.e - r_mat.a * r_mat.g) * r_mat.j \
            + (r_mat.a * r_mat.f - r_mat.b * r_mat.e) * r_mat.k
        if dott < 0:
            scale.x *= -1
            r_mat.a *= -1
            r_mat.b *= -1
            r_mat.c *= -1
        trace = r_mat.a + r_mat.f + r_mat.k
        if trace > 0.00000001:
            s = sqrt(trace + 1.0)
            i_s = 0.5 / s
            rotate = Quaternion(
                (r_mat.g - r_mat.j) * i_s,
                (r_mat.i - r_mat.c) * i_s,
                (r_mat.b - r_mat.e) * i_s,
                s * 0.5
            )
        else:
            if r_mat.a >= r_mat.f and r_mat.a >= r_mat.k:
                s = sqrt(1.0 + r_mat.a - r_mat.f - r_mat.k)
                i_s = 0.5 / s
                rotate = Quaternion(
                    0.5 * s,
                    (r_mat.b + r_mat.e) * i_s,
                    (r_mat.i + r_mat.c) * i_s,
                    (r_mat.g - r_mat.j) * i_s,
                )
            elif r_mat.f > r_mat.k:
                s = sqrt(1.0 + r_mat.f - r_mat.a - r_mat.k)
                i_s = 0.5 / s
                rotate = Quaternion(
                    (r_mat.e + r_mat.b) * i_s,
                    0.5 * s,
                    (r_mat.j + r_mat.g) * i_s,
                    (r_mat.c - r_mat.i) * i_s
                )
            else:
                s = sqrt(1.0 + r_mat.k - r_mat.a - r_mat.f)
                i_s = 0.5 / s
                rotate = Quaternion(
                    (r_mat.c + r_mat.i) * i_s,
                    (r_mat.j + r_mat.g) * i_s,
                    0.5 * s,
                    (r_mat.b - r_mat.e) * i_s
                )

        return translate, rotate, scale


class BoundingSphere:
    __slots__ = ('point', 'radius')
    
    def __init__(self, point: Vector, radius: float):
        self.point = point
        self.radius = radius
    
    def __str__(self):
        p = self.point
        r = self.radius
        return f'{p.x, p.y, p.z} {r}'



class BoundingBox:
    __slots__ = ('min_vec', 'max_vec')
    
    def __init__(self, min_vec: Vector, max_vec: Vector):
        self.min_vec = min_vec
        self.max_vec = max_vec
        
    def __str__(self):
        min = self.min_vec
        max = self.max_vec
        return f'{min.x, min.y, min.z} {max.x, max.y, max.z}'


from math import sqrt
from io import BytesIO
from .stream import BinStream
from .structs import Quaternion, Vector
from .ermmm import Elf


class ANMHepler:
    @staticmethod
    def decompress_quat(bytes):
        first = bytes[0] | (bytes[1] << 8)
        second = bytes[2] | (bytes[3] << 8)
        third = bytes[4] | (bytes[5] << 8)
        bits = first | second << 16 | third << 32
        max_index = (bits >> 45) & 3
        one_div_sqrt2 = 0.70710678118
        sqrt2_div_32767 = 0.00004315969
        a = ((bits >> 30) & 32767) * sqrt2_div_32767 - one_div_sqrt2
        b = ((bits >> 15) & 32767) * sqrt2_div_32767 - one_div_sqrt2
        c = (bits & 32767) * sqrt2_div_32767 - one_div_sqrt2
        d = sqrt(max(0.0, 1.0 - (a * a + b * b + c * c)))
        if max_index == 0:
            return Quaternion(d, a, b, c)
        elif max_index == 1:
            return Quaternion(a, d, b, c)
        elif max_index == 2:
            return Quaternion(a, b, d, c)
        else:
            return Quaternion(a, b, c, d)
        
    @staticmethod
    def compress_quat(quat):
        sqrt_2 = 1.41421356237
        max_index = 3
        abs_x, abs_y, abs_z, abs_w = abs(quat.x), abs(quat.y), abs(quat.z), abs(quat.w)
        if abs_x >= abs_w and abs_x >= abs_y and abs_x >= abs_z:
            max_index = 0
            if quat.x < 0:
                quat *= -1
        elif abs_y >= abs_w and abs_y >= abs_x and abs_y >= abs_z:
            max_index = 1
            if quat.y < 0:
                quat *= -1
        elif abs_z >= abs_w and abs_z >= abs_x and abs_z >= abs_y:
            max_index = 2
            if quat.z < 0:
                quat *= -1
        elif quat.w < 0:
            quat *= -1
        bits = max_index << 45
        quatvalues = (quat.x, quat.y, quat.z, quat.w)
        compressed_index = 0
        for i in range(4):
            if i == max_index:
                continue
            temp = round(16383.5 * (sqrt_2 * quatvalues[i] + 1.0))
            bits |= (temp & 32767) << (30 - 15 * compressed_index)
            compressed_index += 1
        compressed = [(bits >> (8 * i)) & 255 for i in range(6)]
        return bytes(compressed)

    @staticmethod
    def decompress_vec3(min, max, bytes):
        return Vector(
            (max.x - min.x) / 65535.0 * (bytes[0] | bytes[1] << 8) + min.x,
            (max.y - min.y) / 65535.0 * (bytes[2] | bytes[3] << 8) + min.y,
            (max.z - min.z) / 65535.0 * (bytes[4] | bytes[5] << 8) + min.z
        )
        
    @staticmethod
    def evaluate_all_frames(anm):
        # compressed anm need this
        for track in anm.tracks:
            track.poses = dict(sorted(track.poses.items()))
            for frame in range(anm.duration):
                # find left right translate, rotate, scale frames
                left_translate = None
                right_translate = None
                left_scale = None
                right_scale = None
                left_rotate = None
                right_rotate = None
                for time in track.poses.keys():
                    if time <= frame:
                        if track.poses[time].translate != None:
                            left_translate = time
                        if track.poses[time].scale != None:
                            left_scale = time
                        if track.poses[time].rotate != None:
                            left_rotate = time
                    if time >= frame:
                        if track.poses[time].translate != None and right_translate == None:
                            right_translate = time
                        if track.poses[time].scale != None and right_scale == None:
                            right_scale = time
                        if track.poses[time].rotate != None and right_rotate == None:
                            right_rotate = time
                    if left_translate != None and right_translate != None and left_scale != None and right_scale != None and left_rotate != None and right_rotate != None:
                        break
                # create pose if empty
                if frame not in track.poses:
                    track.poses[frame] = pose = ANMPose()
                # evaluate translate if need
                if frame != left_translate and frame != right_translate:
                    if left_translate == None or right_translate == None:
                        if left_translate == None and right_translate == None:
                            raise Exception(
                                f'pyRitoFile: Failed: Write ANM: Evaluate translate: left: {left_translate}, right: {right_translate}.')
                        elif left_translate == None:
                            pose.translate = track.poses[right_translate].translate
                        elif right_translate == None:
                            pose.translate = track.poses[left_translate].translate
                    else:
                        pose.translate = Vector.lerp(
                            track.poses[left_translate].translate, 
                            track.poses[right_translate].translate,
                            (frame - left_translate) / (right_translate - left_translate)
                        )
                # evaluate scale if need
                if frame != left_scale and frame != right_scale:
                    if left_scale == None or right_scale == None:
                        if left_scale == None and right_scale == None:
                            raise Exception(
                                f'pyRitoFile: Failed: Write ANM: Evaluate scale: left: {left_scale}, right: {right_scale}.')
                        elif left_scale == None:
                            pose.scale = track.poses[right_scale].scale
                        elif right_scale == None:
                            pose.scale = track.poses[left_scale].scale
                    else:
                        pose.scale = Vector.lerp(
                            track.poses[left_scale].scale, 
                            track.poses[right_scale].scale,
                            (frame - left_scale) / (right_scale - left_scale)
                        )
                # evaluate rotate if need
                if frame != left_rotate and frame != right_rotate:
                    if left_rotate == None or right_rotate == None:
                        if left_rotate == None and right_rotate == None:
                            raise Exception(
                                f'pyRitoFile: Failed: Write ANM: Evaluate rotate: left: {left_rotate}, right: {right_rotate}.')
                        elif left_rotate == None:
                            pose.rotate = track.poses[right_rotate].rotate
                        elif right_rotate == None:
                            pose.rotate = track.poses[left_rotate].rotate
                    else:
                        pose.rotate = Quaternion.slerp(
                            track.poses[left_rotate].rotate, 
                            track.poses[right_rotate].rotate,
                            (frame - left_rotate) / (right_rotate - left_rotate)
                        )
        
    @staticmethod
    def build_uni_vecs_quats_frames(anm):
        # build uni vecs, uni quats, frame data
        uni_vecs = []
        uni_vec_count = 0
        uni_quats = []
        uni_quat_count = 0
        track_count = len(anm.tracks)
        frames = [None] * anm.duration * track_count
        vec_tol_right = 0.05
        vec_tol_left = -vec_tol_right
        quat_tol_right = 0.005
        quat_tol_left = -quat_tol_right
        for t, track in enumerate(anm.tracks):
            for f in range(anm.duration):
                translate, rotate, scale = track.poses[f].translate, track.poses[f].rotate, track.poses[f].scale
                # translate check
                translate_index = -1
                for vec_index, vec in enumerate(uni_vecs):
                    if vec_tol_left < translate.x-vec.x < vec_tol_right and vec_tol_left < translate.y-vec.y < vec_tol_right and vec_tol_left < translate.z-vec.z < vec_tol_right:
                        translate_index = vec_index
                        break
                if translate_index == -1:
                    uni_vecs.append(translate)
                    translate_index = uni_vec_count
                    uni_vec_count += 1
                # scale check
                scale_index = -1
                for vec_index, vec in enumerate(uni_vecs):
                    if vec_tol_left < scale.x-vec.x < vec_tol_right and vec_tol_left < scale.y-vec.y < vec_tol_right and vec_tol_left < scale.z-vec.z < vec_tol_right:
                        scale_index = vec_index
                        break
                if scale_index == -1:
                    uni_vecs.append(scale)
                    scale_index = uni_vec_count
                    uni_vec_count += 1
                # rotate check
                rotate_index = -1
                for quat_index, quat in enumerate(uni_quats):
                    if quat_tol_left < rotate.x-quat.x < quat_tol_right and quat_tol_left < rotate.y-quat.y < quat_tol_right and quat_tol_left < rotate.z-quat.z < quat_tol_right and quat_tol_left < rotate.w-quat.w < quat_tol_right:
                        rotate_index = quat_index
                        break 
                if rotate_index == -1:
                    uni_quats.append(rotate)
                    rotate_index = uni_quat_count
                    uni_quat_count += 1
                # add to frame
                frames[f * track_count + t] = (translate_index, scale_index, rotate_index)
        return uni_vecs, uni_quats, frames

class ANMErrorMetric:
    __slots__ = (
        'margin', 'discontinuity_threshold'
    )

    def __init__(self, margin=None, discontinuity_threshold=None):
        self.margin = margin
        self.discontinuity_threshold = discontinuity_threshold

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}


class ANMPose:
    __slots__ = (
        'translate', 'rotate', 'scale'
    )

    def __init__(self, translate=None, rotate=None, scale=None):
        self.translate = translate
        self.rotate = rotate
        self.scale = scale

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}


class ANMTrack:
    __slots__ = (
        'joint_hash', 'poses'
    )

    def __init__(self, joint_hash=None, poses=None):
        self.joint_hash = joint_hash
        self.poses = poses # poses[time] = pose at time

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}


class ANM:
    __slots__ = (
        'signature', 'version', 'file_size', 'format_token', 'flags1', 'flags2',
        'duration', 'fps', 'error_metrics', 'tracks'
    )

    def __init__(self, signature=None, version=None, file_size=None, format_token=None, flags1=None, flags2=None, duration=None, fps=None, error_metrics=None, tracks=None):
        self.signature = signature
        self.version = version
        self.file_size = file_size
        self.format_token = format_token
        self.flags1 = flags1
        self.flags2 = flags2
        self.duration = duration
        self.fps = fps
        self.error_metrics = error_metrics
        self.tracks = tracks

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

    def stream(self, path, mode, raw=None):
        if raw != None:
            if raw == True:  # the bool True value
                return BinStream(BytesIO())
            else:
                return BinStream(BytesIO(raw))
        return BinStream(open(path, mode))

    def read(self, path, raw=None):
        with self.stream(path, 'rb', raw) as bs:
            self.signature, = bs.read_s(8)
            self.version, = bs.read_u32()

            if self.signature == 'r3d2canm':
                # compressed
                # read header
                self.file_size, self.format_token, self.flags1 = bs.read_u32(3)
                track_count, frame_count = bs.read_u32(2)
                bs.pad(4)  # jump cache count
                max_time, self.fps = bs.read_f32(2)
                self.duration = (max_time + 1) * self.fps
                # read error metrics
                self.error_metrics = {'rotate': ANMErrorMetric(
                ), 'translate': ANMErrorMetric(), 'scale': ANMErrorMetric()}
                self.error_metrics['rotate'].margin, self.error_metrics['rotate'].discontinuity_threshold = bs.read_f32(
                    2)
                self.error_metrics['translate'].margin, self.error_metrics['translate'].discontinuity_threshold = bs.read_f32(
                    2)
                self.error_metrics['scale'].margin, self.error_metrics['scale'].discontinuity_threshold = bs.read_f32(
                    2)
                translate_min, translate_max, scale_min, scale_max = bs.read_vec3(
                    4)
                # read offsets
                frames_offset,  _, joint_hashes_offset = bs.read_i32(3)
                if frames_offset <= 0:
                    raise Exception(
                        f'pyRitoFile: Failed: Read ANM: File does not contain frames.'
                    )
                if joint_hashes_offset <= 0:
                    raise Exception(
                        f'pyRitoFile: Failed: Read ANM: File does not contain joint hashes.'
                    )
                # read joint hashes
                bs.seek(joint_hashes_offset + 12)
                joint_hashes = bs.read_u32(track_count)
                # create tracks
                self.tracks = [ANMTrack() for i in range(track_count)]
                for track_id, track in enumerate(self.tracks):
                    track.joint_hash = joint_hashes[track_id]
                    track.poses = {}
                # read frames
                bs.seek(frames_offset + 12)
                for i in range(frame_count):
                    compressed_time, bits = bs.read_u16(2)
                    compressed_transform = bs.read(6)
                    # parse track
                    joint_hash = joint_hashes[bits & 16383]
                    match_track = None
                    for track in self.tracks:
                        if track.joint_hash == joint_hash:
                            match_track = track
                            break
                    if match_track == None:
                        # this frame has wrong joint hash?
                        continue
                    # parse pose
                    time = (compressed_time / 65535.0 *
                            max_time) * self.fps
                    if time in match_track.poses:
                        pose = match_track.poses[time]
                    else:
                        pose = ANMPose()
                        match_track.poses[time] = pose
                    # decompress pose data
                    transform_type = bits >> 14
                    if transform_type == 0:
                        pose.rotate = ANMHepler.decompress_quat(
                            compressed_transform)
                    elif transform_type == 1:
                        pose.translate = ANMHepler.decompress_vec3(
                            translate_min, translate_max, compressed_transform)
                    elif transform_type == 2:
                        pose.scale = ANMHepler.decompress_vec3(
                            scale_min, scale_max, compressed_transform)
                    else:
                        raise Exception(
                            f'pyRitoFile: Failed: Read ANM: Unknown compressed transform type: {transform_type}.'
                        )
            elif self.signature == 'r3d2anmd':
                if self.version == 5:
                    # v5
                    # read headers
                    self.file_size, self.format_token, self.flags1, self.flags2 = bs.read_u32(
                        4)
                    track_count, frame_count = bs.read_u32(2)
                    self.fps = 1 / bs.read_f32()[0]  # frame duration
                    self.duration = frame_count
                    # read offsets and calculate stuffs
                    joint_hashes_offset, _, _, vecs_offset, quats_offset, frames_offset = bs.read_i32(
                        6)
                    if joint_hashes_offset <= 0:
                        raise Exception(
                            f'pyRitoFile: Failed: Read ANM: File does not contain joint hashes data.'
                        )
                    if vecs_offset <= 0:
                        raise Exception(
                            f'pyRitoFile: Failed: Read ANM: File does not contain unique vectors data.'
                        )
                    if quats_offset <= 0:
                        raise Exception(
                            f'pyRitoFile: Failed: Read ANM: File does not contain unique quaternions data.'
                        )
                    if frames_offset <= 0:
                        raise Exception(
                            f'pyRitoFile: Failed: Read ANM: File does not contain frames data.'
                        )
                    joint_hash_count = (
                        frames_offset - joint_hashes_offset) // 4
                    vec_count = (quats_offset - vecs_offset) // 12
                    quat_count = (joint_hashes_offset - quats_offset) // 6
                    # read joint hashes
                    bs.seek(joint_hashes_offset + 12)
                    joint_hashes = bs.read_u32(joint_hash_count)
                    # read vecs
                    bs.seek(vecs_offset + 12)
                    uni_vecs = bs.read_vec3(vec_count)
                    # read quats
                    bs.seek(quats_offset + 12)
                    uni_quats = [ANMHepler.decompress_quat(
                        bs.read(6)) for i in range(quat_count)]
                    # prepare tracks
                    self.tracks = [ANMTrack() for i in range(track_count)]
                    for track_id, track in enumerate(self.tracks):
                        track.joint_hash = joint_hashes[track_id]
                        track.poses = {}
                    # read frames
                    bs.seek(frames_offset + 12)
                    for f in range(frame_count):
                        for track in self.tracks:
                            translate_index, scale_index, rotate_index = bs.read_u16(3)
                            # parse pose
                            pose = ANMPose()
                            translate = uni_vecs[translate_index]
                            pose.translate = Vector(
                                translate.x, translate.y, translate.z)
                            scale = uni_vecs[scale_index]
                            pose.scale = Vector(scale.x, scale.y, scale.z)
                            rotate = uni_quats[rotate_index]
                            pose.rotate = Quaternion(
                                rotate.x, rotate.y, rotate.z, rotate.w)
                            track.poses[f] = pose
                    
                elif self.version == 4:
                    # v4
                    # read headers
                    self.file_size, self.format_token, self.flags1, self.flags2 = bs.read_u32(
                        4)
                    track_count, frame_count = bs.read_u32(2)
                    self.fps = 1 / bs.read_f32()[0]  # frame duration
                    self.duration = frame_count
                    # read offsets & calculate stuffs
                    _, _, _, vecs_offset, quats_offset, frames_offset = bs.read_i32(
                        6)
                    if vecs_offset <= 0:
                        raise Exception(
                            f'pyRitoFile: Failed: File does not contain unique vectors data.'
                        )
                    if quats_offset <= 0:
                        raise Exception(
                            f'pyRitoFile: Failed: File does not contain unique quaternions data.'
                        )
                    if frames_offset <= 0:
                        raise Exception(
                            f'pyRitoFile: Failed: File does not contain frames data.'
                        )
                    vec_count = (quats_offset - vecs_offset) // 12
                    quat_count = (frames_offset - quats_offset) // 16
                    # read uni vecs
                    bs.seek(vecs_offset + 12)
                    uni_vecs = bs.read_vec3(vec_count)
                    # read uni quats
                    bs.seek(quats_offset + 12)
                    uni_quats = bs.read_quat(quat_count)
                    # prepare tracks
                    self.tracks = [ANMTrack() for i in range(track_count)]
                    for track in self.tracks:
                        track.poses = {}
                    # read frames
                    bs.seek(frames_offset + 12)
                    for f in range(frame_count):
                        for t in range(track_count):
                            joint_hash, = bs.read_u32()
                            translate_index, scale_index, rotate_index = bs.read_u16(3)
                            bs.pad(2)
                            track = self.tracks[t]
                            if track.joint_hash == None:
                                # if track t is new
                                track.joint_hash = joint_hash
                            else:
                                # if track t is not new
                                if track.joint_hash != joint_hash:
                                    # track t is wrong for this joint hash, find another track
                                    match_track = None
                                    for track in self.tracks:
                                        if track.joint_hash == joint_hash:
                                            match_track = track
                                            break
                                    if match_track == None:
                                        # no track found???
                                        continue
                            # parse pose
                            pose = ANMPose()
                            translate = uni_vecs[translate_index]
                            pose.translate = Vector(
                                translate.x, translate.y, translate.z)
                            scale = uni_vecs[scale_index]
                            pose.scale = Vector(scale.x, scale.y, scale.z)
                            rotate = uni_quats[rotate_index]
                            pose.rotate = Quaternion(
                                rotate.x, rotate.y, rotate.z, rotate.w)
                            track.poses[len(track.poses)] = pose
                elif self.version == 3:
                    # legacy
                    # read headers
                    bs.pad(4)  # skl id
                    track_count, frame_count = bs.read_u32(2)
                    self.fps, = bs.read_u32()
                    self.duration = frame_count
                    # prepare tracks
                    self.tracks = [ANMTrack() for i in range(track_count)]
                    for track in self.tracks:
                        track.joint_hash = Elf(bs.read_s_padded(32)[0])
                        bs.pad(4)  # flags
                        # read pose
                        track.poses = {}
                        for f in range(frame_count):
                            pose = ANMPose()
                            pose.rotate, = bs.read_quat()
                            pose.translate, = bs.read_vec3()
                            # legacy not support scaling
                            pose.scale = Vector(1.0, 1.0, 1.0)
                            track.poses[f] = pose
                else:
                    raise Exception(
                        f'pyRitoFile: Failed: Read ANM: Unsupported file version: {self.version}')
            else:
                raise Exception(
                    f'pyRitoFile: Failed: Read ANM: Wrong signature file: {hex(self.signature)}') 

    def write(self, path, raw=None):
        with self.stream(path, 'wb', raw) as bs:
            self.duration = int(self.duration)
            ANMHepler.evaluate_all_frames(self)
            uni_vecs, uni_quats, frames = ANMHepler.build_uni_vecs_quats_frames(self)
            # start write anm
            bs.write_s('r3d2anmd') # signature
            bs.write_u32(
                5, # version
                0, # file_size
                0, # format token
                0, # flags1
                0, # flags2
            )
            bs.write_u32(
                len(self.tracks), # track_count
                self.duration # frame_count
            )
            bs.write_f32(1 / self.fps) # fps
            bs.write_u32(
                0, # joint_hashes_offset
                0,
                0,
                0, # vecs_offset
                0, # quats_offset
                0 # frames_offset
            )
            bs.write_u32(0, 0, 0) # pad 12 bytes
            # write data 
            # must in order: vecs -> quats -> joint_hashses -> frames
            # vec
            vecs_offset = bs.tell()
            bs.write_vec3(*uni_vecs)
            # quat
            quat_offsets = bs.tell()
            bs.write(b''.join(ANMHepler.compress_quat(quat) for quat in uni_quats))
            # joint_hash
            joint_hashes_offset = bs.tell()
            bs.write_u32(*[track.joint_hash for track in self.tracks])
            # frame
            frames_offset = bs.tell()
            for translate_index, rotate_index, scale_index in frames:
                bs.write_u16(translate_index, rotate_index, scale_index)
            # write offsets
            bs.seek(12) 
            bs.write_u32(bs.end()) # file_size
            bs.seek(40)
            bs.write_u32(joint_hashes_offset-12)
            bs.seek(52)
            bs.write_u32(vecs_offset-12, quat_offsets-12, frames_offset-12)
            return bs.raw() if raw else None


from io import BytesIO
from .stream import BinStream
from .ermmm import FNV1a
from enum import IntEnum

def hash_to_hex(hash):
    return f'{hash:08x}'


def hex_to_hash(hex):
    return int(hex, 16)


def name_to_hash(name):
    return FNV1a(name)


def hex_to_name(hashtables, table_name, hash):
    return hashtables.get(table_name, {}).get(hash, hash)


def name_to_hex(name):
    return hash_to_hex(name_to_hash(name))


def name_or_hex_to_hash(value):
    try:
        return hex_to_hash(value)
    except:
        return name_to_hash(value)

class BINType(IntEnum):
    # basic
    Empty = 0
    Bool = 1
    I8 = 2
    U8 = 3
    I16 = 4
    U16 = 5
    I32 = 6
    U32 = 7
    I64 = 8
    U64 = 9
    F32 = 10
    Vec2 = 11
    Vec3 = 12
    Vec4 = 13
    Mtx4 = 14
    RGBA = 15
    String = 16
    Hash = 17
    File = 18
    # complex
    List = 128
    List2 = 129
    Pointer = 130
    Embed = 131
    Link = 132
    Option = 133
    Map = 134
    Flag = 135

    def __json__(self):
        return self.name

class BINHelper:
    legacy_read = False

    @staticmethod
    def fix_type(bin_type):
        if BINHelper.legacy_read:
            if bin_type >= 129:
                bin_type += 1
        return BINType(bin_type)

    @staticmethod
    def find_item(*, items=[], compare_func=None, return_func=None):
        if compare_func != None and len(items) > 0:
            for item in items:
                if compare_func(item):
                    if return_func != None:
                        return return_func(item)
                    else:
                        return item
        return None
    
    # read related stuffs
    read_value_dict = {
        BINType.Empty:      lambda bs: bs.read_u16(3),
        BINType.Bool:       lambda bs: bs.read_b()[0],
        BINType.I8:         lambda bs: bs.read_i8()[0],
        BINType.U8:         lambda bs: bs.read_u8()[0],
        BINType.I16:        lambda bs: bs.read_i16()[0],
        BINType.U16:        lambda bs: bs.read_u16()[0],
        BINType.I32:        lambda bs: bs.read_i32()[0],
        BINType.U32:        lambda bs: bs.read_u32()[0],
        BINType.I64:        lambda bs: bs.read_i64()[0],
        BINType.U64:        lambda bs: bs.read_u64()[0],
        BINType.F32:        lambda bs: bs.read_f32()[0],
        BINType.Vec2:       lambda bs: bs.read_vec2()[0],
        BINType.Vec3:       lambda bs: bs.read_vec3()[0],
        BINType.Vec4:       lambda bs: bs.read_vec4()[0],
        BINType.Mtx4:       lambda bs: bs.read_mtx4()[0],
        BINType.RGBA:       lambda bs: bs.read_u8(4),
        BINType.String:     lambda bs: bs.read_s_sized16(encoding='utf-8')[0],
        BINType.Hash:       lambda bs: hash_to_hex(bs.read_u32()[0]),
        BINType.File:       lambda bs: bs.read_u64()[0],
        BINType.Pointer:    lambda bs: BINHelper.read_pointer_or_embed(bs, BINField(type=BINType.Pointer)),
        BINType.Embed:      lambda bs: BINHelper.read_pointer_or_embed(bs, BINField(type=BINType.Embed)),
        BINType.Link:       lambda bs: hash_to_hex(bs.read_u32()[0]),
        BINType.Flag:       lambda bs: bs.read_u8()[0],
    }

    @staticmethod
    def read_value(bs, value_type):
        return BINHelper.read_value_dict[value_type](bs)

    @staticmethod
    def read_basic(bs, field):
        field.data = BINHelper.read_value(bs, field.type)
        return field

    @staticmethod
    def read_list_or_list2(bs, field):
        field.value_type = BINHelper.fix_type(bs.read_u8()[0])
        bs.pad(4)  # size
        count, = bs.read_u32()
        field.data = [
            BINHelper.read_value(bs, field.value_type)
            for i in range(count)
        ]
        return field
        
    @staticmethod
    def read_pointer_or_embed(bs, field):
        field.hash_type = hash_to_hex(bs.read_u32()[0])
        if field.hash_type != '00000000':
            bs.pad(4)  # size
            count, = bs.read_u16()
            field.data = [
                BINHelper.read_field(bs)
                for i in range(count)
            ]
        else:
            field.data = None
        return field
    
    @staticmethod
    def read_option(bs, field):
        field.value_type = BINHelper.fix_type(bs.read_u8()[0])
        count, = bs.read_u8()
        if count != 0:
            field.data = BINHelper.read_value(bs, field.value_type)
        else:
            field.data = None
        return field

    @staticmethod
    def read_map(bs, field):
        field.key_type = BINHelper.fix_type(bs.read_u8()[0])
        field.value_type = BINHelper.fix_type(bs.read_u8()[0])
        bs.pad(4)  # size
        count, = bs.read_u32()

        # Boring, old, threating map as key: value, so it was deleting stuff that was copy-pasted with same name
        #field.data = {
        #    BINHelper.read_value(bs, field.key_type): BINHelper.read_value(bs, field.value_type)
        #    for i in range(count)
        #}

        # New, chad, list of key and value pairs
        field.data = [
            (BINHelper.read_value(bs, field.key_type), BINHelper.read_value(bs, field.value_type))
            for _ in range(count)
        ]

        return field
    
    read_field_dict = {
        field_type:         lambda bs, field: BINHelper.read_basic(bs, field) 
        for field_type in read_value_dict
    }
    read_field_dict.update({
        BINType.List:       lambda bs, field: BINHelper.read_list_or_list2(bs, field),
        BINType.List2:      lambda bs, field: BINHelper.read_list_or_list2(bs, field),
        BINType.Pointer:    lambda bs, field: BINHelper.read_pointer_or_embed(bs, field),
        BINType.Embed:      lambda bs, field: BINHelper.read_pointer_or_embed(bs, field),
        BINType.Option:     lambda bs, field: BINHelper.read_option(bs, field),
        BINType.Map:        lambda bs, field: BINHelper.read_map(bs, field)
    })

    @staticmethod
    def read_field(bs):
        field = BINField(
            hash=hash_to_hex(bs.read_u32()[0]),
            type=BINHelper.fix_type(bs.read_u8()[0])
        )
        return BINHelper.read_field_dict[field.type](bs, field)

    # write related stuffs
    write_value_dict = {
        BINType.Empty:          lambda bs, value: (bs.write_u16(*value), 0),
        BINType.Bool:           lambda bs, value: (bs.write_b(value), 1),
        BINType.I8:             lambda bs, value: (bs.write_i8(value), 1),
        BINType.U8:             lambda bs, value: (bs.write_u8(value), 1),
        BINType.I16:            lambda bs, value: (bs.write_i16(value), 2),
        BINType.U16:            lambda bs, value: (bs.write_u16(value), 2),
        BINType.I32:            lambda bs, value: (bs.write_i32(value), 4),
        BINType.U32:            lambda bs, value: (bs.write_u32(value), 4),
        BINType.I64:            lambda bs, value: (bs.write_i64(value), 8),
        BINType.U64:            lambda bs, value: (bs.write_u64(value), 8),
        BINType.F32:            lambda bs, value: (bs.write_f32(value), 4),
        BINType.Vec2:           lambda bs, value: (bs.write_vec2(value), 8),
        BINType.Vec3:           lambda bs, value: (bs.write_vec3(value), 12),
        BINType.Vec4:           lambda bs, value: (bs.write_vec4(value), 16),
        BINType.Mtx4:           lambda bs, value: (bs.write_mtx4(value), 64),
        BINType.RGBA:           lambda bs, value: (bs.write_u8(*value), 4),
        BINType.String:         lambda bs, value: (bs.write_s_sized16(value, encoding='utf-8'), len(value.encode('utf-8'))+2),
        BINType.Hash:           lambda bs, value: (bs.write_u32(name_or_hex_to_hash(value)), 4),
        BINType.File:           lambda bs, value: (bs.write_u64(value), 8),
        BINType.Pointer:        lambda bs, value: BINHelper.write_pointer_or_embed(bs, value),
        BINType.Embed:          lambda bs, value: BINHelper.write_pointer_or_embed(bs, value),
        BINType.Link:           lambda bs, value: (bs.write_u32(name_or_hex_to_hash(value)), 4),
        BINType.Flag:           lambda bs, value: (bs.write_u8(value), 1),
    }

    @staticmethod
    def write_value(bs, value, value_type, header_size):
        size = BINHelper.write_value_dict[value_type](bs, value)[1]
        return size+5 if header_size else size
    
    @staticmethod
    def write_basic(bs, field):
        return None, BINHelper.write_value(bs, field.data, field.type, header_size=False)

    @staticmethod
    def write_list_or_list2(bs, field):
        size = 0
        bs.write_u8(field.value_type.value)

        return_offset = bs.tell()
        bs.write_u32(0)  # values size
        size += 1 + 4

        content_size = 4
        bs.write_u32(len(field.data))
        for value in field.data:
            content_size += BINHelper.write_value(bs,
                                                    value, field.value_type, header_size=False)
        bs.size_offsets.append((return_offset, content_size))

        size += content_size
        return None, size

    @staticmethod
    def write_pointer_or_embed(bs, field):
        size = 0
        if field.hash_type == '00000000':
            bs.write_u32(0)
            size += 4
        else:
            bs.write_u32(name_or_hex_to_hash(field.hash_type))
            size += 4

            return_offset = bs.tell()
            bs.write_u32(0)  # size
            size += 4

            content_size = 2
            bs.write_u16(len(field.data))
            for value in field.data:
                content_size += BINHelper.write_field(
                    bs, value, header_size=True)
            bs.size_offsets.append((return_offset, content_size))

            size += content_size
        return None, size
    
    @staticmethod
    def write_option(bs, field):
        size = 0
        bs.write_u8(field.value_type.value)
        count = 0 if field.data == None else 1
        bs.write_u8(count)
        size += 1 + 1
        if count != 0:
            size += BINHelper.write_value(bs, field.data, field.value_type, header_size=False)
        return None, size
    
    @staticmethod
    def write_map(bs, field: 'BINField'):
        size = 0
        bs.write_u8(
            field.key_type.value,
            field.value_type.value
        )

        return_offset = bs.tell()
        bs.write_u32(0)  # size
        size += 1+1+4

        content_size = 4
        bs.write_u32(len(field.data))
        for key, value in field.data:
            content_size += BINHelper.write_value(bs,
                                                    key, field.key_type, header_size=False)
            content_size += BINHelper.write_value(bs,
                                                    value, field.value_type, header_size=False)
        bs.size_offsets.append((return_offset, content_size))

        size += content_size
        return None, size
    
    write_field_dict = {
        field_type:         lambda bs, field: BINHelper.write_basic(bs, field) 
        for field_type in write_value_dict
    }
    write_field_dict.update({
        BINType.List:       lambda bs, field: BINHelper.write_list_or_list2(bs, field),
        BINType.List2:      lambda bs, field: BINHelper.write_list_or_list2(bs, field),
        BINType.Pointer:    lambda bs, field: BINHelper.write_pointer_or_embed(bs, field),
        BINType.Embed:      lambda bs, field: BINHelper.write_pointer_or_embed(bs, field),
        BINType.Option:     lambda bs, field: BINHelper.write_option(bs, field),
        BINType.Map:        lambda bs, field: BINHelper.write_map(bs, field)
    })

    @staticmethod
    def write_field(bs, field, header_size):
        bs.write_u32(name_or_hex_to_hash(field.hash))
        bs.write_u8(field.type.value)
        size = BINHelper.write_field_dict[field.type](bs, field)[1]
        return size+5 if header_size else size


class BINField:
    __slots__ = ('hash', 'type', 'hash_type', 'key_type', 'value_type', 'data')

    def __init__(self, hash=None, type=None, hash_type=None, key_type=None, value_type=None, data=None):
        self.hash = hash
        self.type = type
        self.hash_type = hash_type
        self.key_type = key_type
        self.value_type = value_type
        self.data = data

    def __json__(self):
        dic = {key: getattr(self, key) for key in self.__slots__}
        if self.type == BINType.List or self.type == BINType.List2:
            dic.pop('key_type')
            dic.pop('hash_type')
        elif self.type == BINType.Pointer or self.type == BINType.Embed:
            dic.pop('key_type')
            dic.pop('value_type')
        elif self.type == BINType.Map:
            dic.pop('hash_type')
        else:
            dic.pop('key_type')
            dic.pop('hash_type')
            dic.pop('value_type')
        return dic


class BINPatch:
    __slots__ = ('hash', 'path', 'type', 'data')

    def __init__(self, hash=None, path=None, type=None, data=None):
        self.hash = hash
        self.path = path
        self.type = type
        self.data = data

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}


class BINEntry:
    __slots__ = ('hash', 'type', 'data')

    def __init__(self, hash=None, type=None, data=None):
        self.hash = hash
        self.type = type
        self.data = data

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}


class BIN:
    __slots__ = (
        'signature', 'version', 'is_patch',
        'links', 'entries', 'patches'
    )

    def __init__(self, signature=None, version=None, is_patch=False, links=None, entries=None, patches=None):
        self.signature = signature
        self.version = version
        self.is_patch = is_patch
        self.links = links
        self.entries: list[BINEntry] = entries
        self.patches = patches

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

    def stream(self, path, mode, raw=None):
        if raw != None:
            if raw == True:  # the bool True value
                return BinStream(BytesIO())
            else:
                return BinStream(BytesIO(raw))
        return BinStream(open(path, mode))

    def read(self, path, raw=None):
        with self.stream(path, 'rb', raw) as bs:
            # header
            self.signature, = bs.read_s(4, encoding='utf-8')
            if self.signature not in ('PROP', 'PTCH'):
                raise Exception(
                    f'pyRitoFile: Failed: Read BIN {path}: Wrong file signature: {self.signature}')
            if self.signature == 'PTCH':
                self.is_patch = True
                bs.pad(8)  # patch header
                magic, = bs.read_s(4, encoding='utf-8')
                if magic != 'PROP':
                    raise Exception(
                        f'pyRitoFile: Failed: Read BIN {path}: Missing PROP after PTCH signature.')
            self.version, = bs.read_u32()
            if self.version not in (1, 2, 3):
                raise Exception(
                    f'pyRitoFile: Failed: Read BIN {path}: Unsupported file version: {self.version}')
            # links
            if self.version >= 2:
                link_count, = bs.read_u32()
                self.links = [bs.read_s_sized16(encoding='utf-8')[0] for _ in range(link_count)]
            # entry_types + entries
            entry_count, = bs.read_u32()
            entry_types = bs.read_u32(entry_count)
            entry_offset = bs.tell()
            try:
                # read as new bin
                self.entries = [BINEntry() for i in range(entry_count)]
                for entry_id, entry in enumerate(self.entries):
                    entry.type = hash_to_hex(entry_types[entry_id])
                    bs.pad(4)  # size
                    entry.hash = hash_to_hex(bs.read_u32()[0])
                    field_count, = bs.read_u16()
                    entry.data = [BINHelper.read_field(
                        bs) for i in range(field_count)]
            except ValueError:
                # legacy bin, fall back
                bs.seek(entry_offset)
                BINHelper.legacy_read = True
                self.entries = [BINEntry() for i in range(entry_count)]
                for entry_id, entry in enumerate(self.entries):
                    entry.type = hash_to_hex(entry_types[entry_id])
                    bs.pad(4)  # size
                    entry.hash = hash_to_hex(bs.read_u32()[0])
                    field_count, = bs.read_u16()
                    entry.data = [BINHelper.read_field(
                        bs) for i in range(field_count)]
                BINHelper.legacy_read = False
            except Exception as e:
                # raise any other errors
                raise e
            # patches
            if self.is_patch and self.version >= 3:
                patch_count, = bs.read_u32()
                self.patches = [BINPatch() for i in range(patch_count)]
                for patch in self.patches:
                    patch.hash = hash_to_hex(bs.read_u32()[0])
                    bs.pad(4)  # size
                    patch.type = BINHelper.fix_type(bs.read_u8()[0])
                    patch.path, = bs.read_s_sized16(encoding='utf-8')
                    patch.data = BINHelper.read_value(bs, patch.type)

    def write(self, path, raw=None):
        with self.stream(path, 'wb', raw) as bs:
            # header
            if self.is_patch:
                bs.write_s('PTCH', encoding='utf-8')
                bs.write_u32(1, 0)  # patch header
            bs.write_s('PROP', encoding='utf-8')
            bs.write_u32(3)  # version
            # links
            bs.write_u32(len(self.links))
            for link in self.links:
                bs.write_s_sized16(link, encoding='utf-8')
            # entry_types + entries
            bs.write_u32(len(self.entries))
            for entry in self.entries:
                bs.write_u32(name_or_hex_to_hash(entry.type))
            
            bs.size_offsets = []  # this help to write sizes
            for entry in self.entries:
                return_offset = bs.tell()

                bs.write_u32(0)  # size
                entry_size = 4+2

                bs.write_u32(name_or_hex_to_hash(entry.hash))
                bs.write_u16(len(entry.data))
                for field in entry.data:
                    entry_size += BINHelper.write_field(
                        bs, field, header_size=True)
                bs.size_offsets.append((return_offset, entry_size))
            # patches
            if self.is_patch:
                bs.write_u32(len(self.patches))
                for patch in self.patches:
                    bs.write_u32(name_or_hex_to_hash(path.hash))

                    return_offset = bs.tell()
                    bs.write_u32(0)  # size
                    patch_size = 1 + 2 + len(patch.path)

                    bs.write_u8(patch.type.value)
                    bs.write_s_sized16(patch.path, encoding='utf-8')
                    patch_size += BINHelper.write_value(
                        bs, patch.type, header_size=False)
                    bs.size_offsets.append(
                        (return_offset, patch_size))
            # jump around and write size
            for offset, size in bs.size_offsets:
                bs.seek(offset)
                bs.write_u32(size)
            return bs.raw() if raw else None

    def un_hash(self, hashtables=None):
        if hashtables == None:
            return

        def un_hash_value(value, value_type):
            if value_type == BINType.Hash:
                return hex_to_name(hashtables, 'hashes.binhashes.txt', value)
            elif value_type == BINType.Link:
                return hex_to_name(hashtables, 'hashes.binentries.txt', value)
            elif value_type in (BINType.List, BINType.List2):
                value.data = [un_hash_value(v, value_type) for v in value.data]
            elif value_type in (BINType.Embed, BINType.Pointer):
                if value.hash_type != '00000000':
                    value.hash_type = hex_to_name(
                        hashtables, 'hashes.bintypes.txt',  value.hash_type)
                    for f in value.data:
                        un_hash_field(f)
            return value

        def un_hash_field(field):
            field.hash = hex_to_name(
                hashtables, 'hashes.binfields.txt', field.hash)
            field.type = hex_to_name(
                hashtables, 'hashes.bintypes.txt', field.type)
            if field.type in (BINType.List, BINType.List2):
                field.value_type = hex_to_name(
                    hashtables, 'hashes.bintypes.txt', field.value_type)
                field.data = [un_hash_value(v, field.value_type)
                              for v in field.data]
            elif field.type in (BINType.Embed, BINType.Pointer):
                if field.hash_type != '00000000':
                    field.hash_type = hex_to_name(
                        hashtables, 'hashes.bintypes.txt', field.hash_type)
                    for f in field.data:
                        un_hash_field(f)
            elif field.type == BINType.Map:
                field.key_type = hex_to_name(
                    hashtables, 'hashes.bintypes.txt', field.key_type)
                field.value_type = hex_to_name(
                    hashtables, 'hashes.bintypes.txt', field.value_type)
                field.data = {
                    un_hash_value(key, field.key_type): un_hash_value(value, field.value_type) for key, value in field.data.items()
                }
            else:
                field.data = un_hash_value(field.data, field.type)

        for entry in self.entries:
            entry.hash = hex_to_name(
                hashtables, 'hashes.binentries.txt', entry.hash)
            entry.type = hex_to_name(
                hashtables, 'hashes.bintypes.txt', entry.type)
            for field in entry.data:
                un_hash_field(field)



from io import BytesIO
from typing import Optional
from .stream import BinStream
from .structs import Matrix4, Quaternion, Vector
from .ermmm import Elf, FNV1a


def bin_hash(name):
    return f'{FNV1a(name):08x}'


class SKLJoint:
    """
    Represents one single joint.

    Fields:
    -------
        `id`: A integer that represents the id of the joint.\n
        `name`: A string that has the name of the joint.\n
        `bin_hash`: A string, name of the joint as FNV1a lowered hash.\n
        `parent`: A integer that points to the parent bone, -1 means world parent (not a child of any joint).\n
        `hash`: A integer that represents the hash from the curret joint.\n
        `radius`: A float that represent the radius of the joint.\n
        `flags`: A integer that represents the flags of the current joint.\n\n
        (GuiSai note: I dont know 3D Stuff)\n
        Lets assume that local stands for the local matrix of the joint.
        And the ibind stands for the inverse bind pose matrix.\n
        `local_translate`: :class:`Vector`.\n
        `local_rotate`: :class:`Quaternion`.\n
        `local_scale`: :class:`Vector`.\n
        `ibind_translate`: :class:`Vector`.\n
        `ibind_rotate`: :class:`Quaternion`.\n
        `ibind_scale`: :class:`Vector`.\n
    """
    
    __slots__ = (
        'id', 'name', 'bin_hash', 'parent', 'hash', 'radius', 'flags',
        'local_translate', 'local_rotate', 'local_scale',
        'ibind_translate', 'ibind_rotate', 'ibind_scale'
    )

    def __init__(self):
        self.id: Optional[int] = None
        self.name: Optional[str] = None
        self.bin_hash: Optional[str] = None
        self.parent: Optional[int] = None
        self.hash: Optional[int] = None
        self.radius: Optional[float] = None
        self.flags: Optional[int] = None
        self.local_translate: Optional[Vector] = None
        self.local_rotate: Optional[Quaternion] = None
        self.local_scale: Optional[Vector] = None
        self.ibind_translate = None
        self.ibind_rotate = None
        self.ibind_scale = None

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}


class SKL:
    """
    Represents a Skeleton from League of Legends, use .read() to get the fields initialized.

    Fields:
    -------\n
        `Unique from new SKL`:
            `flags`: Integer that represents the flags.\n
            `name`: String that represents the name.\n
            `asset`: String that represents the asset.\n

        `Persistents`:
            `signature`: A string that contains the hex value of the SKL signature.\n
            `version`: A integer that represents the version.\n
            `file_size`: A integer that represents the file size.\n
            `joints`: A list of :class:`SKLJoint`.\n
            `influences`: A tuple of integer that represents each unique bound joint id.
            Note: not every joint is inside unfluences, because not every joint have skin bound, for example vfx bones.\n
    
    Methods:
    -------
        `read()`: Used to read any supported SKL file, it fills the fields of the instance.\n
        `write()`: Used to write the current instance.\n
        `stream()`: Used to manually write / read SKL files.\n
        `__json__()`: Returns a dict of every fields. (field: value of the field).\n
    """
    __slots__ = (
        'file_size', 'signature', 'version', 'flags',
        'name', 'asset', 'joints', 'influences'
    )

    def __init__(self):
        self.file_size: Optional[int] = None
        self.signature: Optional[str] = None
        self.version: Optional[int] = None
        self.flags: Optional[int] = None
        self.name: Optional[str] = None
        self.asset: Optional[str] = None
        self.joints: list[SKLJoint] = []
        self.influences: tuple[int, ...] = []  # type: ignore[assignment]
    
    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

    def stream(self, path, mode, raw=None):
        if raw != None:
            if raw == True:  # the bool True value
                return BinStream(BytesIO())
            else:
                return BinStream(BytesIO(raw))
        return BinStream(open(path, mode))

    def read(self, path, raw=None):
        with self.stream(path, 'rb', raw) as bs:
            # read signature first to check legacy or not
            bs.pad(4)
            signature, = bs.read_u32()
            bs.seek(0)

            if signature == 0x22FD4FC3:
                # new skl data
                self.file_size, self.signature, self.version, = bs.read_u32(
                    3)
                if self.version != 0:
                    raise Exception(
                        f'pyRitoFile: Failed: Read SKL {path}: Unsupported file version: {self.version}')
                self.signature = hex(self.signature)
                # unknown
                self.flags, = bs.read_u16()
                # counts
                joint_count, = bs.read_u16()
                influence_count, = bs.read_u32()
                # offsets
                joints_offset, _, influences_offset, name_offset, asset_offset, _, = bs.read_i32(
                    6)
                # pad empty
                bs.pad(20)

                # read joints
                if joints_offset > 0 and joint_count > 0:
                    bs.seek(joints_offset)
                    self.joints = [SKLJoint() for i in range(joint_count)]
                    for joint in self.joints:
                        joint.flags, = bs.read_u16()
                        joint.id, joint.parent, = bs.read_i16(2)
                        bs.pad(2)  # pad
                        joint.hash, = bs.read_u32()
                        joint.radius, = bs.read_f32()
                        joint.local_translate, = bs.read_vec3()
                        joint.local_scale, = bs.read_vec3()
                        joint.local_rotate, = bs.read_quat()

                        joint.ibind_translate, = bs.read_vec3()
                        joint.ibind_scale, = bs.read_vec3()
                        joint.ibind_rotate, = bs.read_quat()
                        # name
                        joint_name_offset, = bs.read_i32()
                        return_offset = bs.tell()
                        bs.seek(return_offset-4 + joint_name_offset)
                        joint.name, = bs.read_c_until0()
                        joint.bin_hash = bin_hash(joint.name)
                        bs.seek(return_offset)

                # read influences
                if influences_offset > 0 and influence_count > 0:
                    bs.seek(influences_offset)
                    self.influences = bs.read_u16(influence_count)
                # name and asset string
                if name_offset > 0:
                    self.name, = bs.read_c_until0()
                if asset_offset > 0:
                    self.asset, = bs.read_c_until0()
            else:
                self.signature, = bs.read_s(8)
                if self.signature != 'r3d2sklt':
                    raise Exception(
                        f'pyRitoFile: Failed: Read SKL {path}: Wrong file signature: {self.signature}')

                self.version, = bs.read_u32()
                if self.version not in (1, 2):
                    raise Exception(
                        f'pyRitoFile: Failed: Read SKL {path}: Unsupported file version: {self.version}')

                # skeleton id
                skeleton_id, = bs.read_u32()

                joint_count, = bs.read_u32()
                self.joints = [SKLJoint() for i in range(joint_count)]
                old_matrices = [None] * joint_count
                for joint_id, joint in enumerate(self.joints):
                    joint.name, = bs.read_s_padded(32)
                    joint.bin_hash = bin_hash(joint.name)
                    joint.id = joint_id
                    joint.hash = Elf(joint.name)
                    joint.parent, = bs.read_i32()
                    joint.radius, = bs.read_f32()
                    floats = [0.0]*16
                    for c in range(3):
                        for r in range(4):
                            floats[r*4+c], = bs.read_f32()
                    floats[15] = 1.0
                    old_matrices[joint_id] = Matrix4(*floats)

                # old matrix to local translation, ibind rotation
                for joint_id, joint in enumerate(self.joints):
                    local = old_matrices[joint_id] if joint.parent == - \
                        1 else old_matrices[joint_id] * old_matrices[joint.parent].inverse()
                    joint.local_translate, joint.local_rotate, joint.local_scale = local.decompose()
                    ibind = old_matrices[joint_id].inverse()
                    joint.ibind_translate, joint.ibind_rotate, joint.ibind_scale = ibind.decompose()

                # read influences
                if self.version == 1:
                    self.influences = list(range(joint_count))

                if self.version == 2:
                    influence_count, = bs.read_u32()
                    self.influences = bs.read_u32(influence_count)

    def write(self, path, raw=None):
        with self.stream(path, 'wb', raw) as bs:
            # file size, magic, version
            bs.write_u32(0, 0x22FD4FC3, 0)

            joint_count = len(self.joints)
            bs.write_u16(0)  # flags
            bs.write_u16(joint_count)
            bs.write_u32(joint_count)

            joints_offset = 64
            joint_indices_offset = joints_offset + joint_count * 100
            influences_offset = joint_indices_offset + joint_count * 8
            joint_names_offset = influences_offset + joint_count * 2

            bs.write_i32(
                joints_offset,
                joint_indices_offset,
                influences_offset,
                0,  # name offset
                0,  # asset offset
                joint_names_offset
            )

            bs.write_u32(0xFFFFFFFF, 0xFFFFFFFF,
                         0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF)

            joint_offset = [None] * joint_count
            bs.seek(joint_names_offset)
            for i in range(joint_count):
                joint_offset[i] = bs.tell()
                bs.write_s(self.joints[i].name)
                bs.write_b(0)

            bs.seek(joints_offset)
            for joint_id, joint in enumerate(self.joints):
                bs.write_u16(0, joint_id)  # flags + id
                bs.write_i16(joint.parent)  # -1, cant be uint
                bs.write_u16(0)  # pad
                bs.write_u32(joint.hash)
                bs.write_f32(joint.radius)

                # local
                bs.write_vec3(joint.local_translate)
                bs.write_vec3(joint.local_scale)
                bs.write_quat(joint.local_rotate)

                # inversed bind
                bs.write_vec3(joint.ibind_translate)
                bs.write_vec3(joint.ibind_scale)
                bs.write_quat(joint.ibind_rotate)

                bs.write_u32(joint_offset[joint_id] - bs.tell())

            # influences
            bs.seek(influences_offset)
            bs.write_u16(*[i for i in range(joint_count)])

            # joint indices
            bs.seek(joint_indices_offset)
            for i in range(joint_count):
                bs.write_u16(i)
                bs.write_u16(0)  # pad
                bs.write_u32(joint.hash)

            # file size
            bs.seek(0)
            bs.write_u32(bs.end())
            return bs.raw() if raw else None



from io import BytesIO, StringIO
from typing import Optional
from .stream import BinStream
from .structs import BoundingBox, Vector
from enum import IntEnum

class SOFlag(IntEnum):
    HasVcp = 1
    HasLocalOriginLocatorAndPivot = 2

    def __json__(self):
        return self.name


class SO:
    """
    Represents either a SCO or an SCB.

    Fields:
    --------
        `Optionals for SCO`:
            `pivot`: :class:`Vector` that represents the pivot of the SCO, because them does not store bones.
        
        `Optionals for SCB`:
            `flags`: A :class:`SOFlag`.
            `version`: A float.
            `vertex_type`: A integer that represents the vertex type.\n
            `bounding_box`: :class:`BoundingBox` that has a min and max vectors.\n
            `colors`: A list of RGBA values, only created if the SCB vertex_type >= 1.\n

        `Shared`:
            `material`: A string that represents the material name from the SO.\n
            `name`: A string that represents the name of the SO.\n
            `indices`: A list of integers that contains the vertices indices.\n
            `positions`: A list of vertices :class:`Vector`.\n
            `central`: A :class:`Vector`.
    
    Methods:
    -------
        `read_sco()`: Used to read any supported SCO file, it fills the fields of the instance.\n
        `read_scb()`: Used to read any supported SCB file, it fills the fields of the instance.\n
        `stream_sco()`: Used to manually write / read SCO files.\n
        `stream_scb()`: Used to manually write / read SCB files.\n
        `__json__()`: Returns a dict of every fields. {field: value of the field}.\n
    """

    __slots__ = (
        'signature', 'version', 'flags', 'name',
        'central', 'pivot', 'bounding_box', 'material', 'vertex_type',
        'indices', 'positions', 'uvs', 'colors'
    )

    def __init__(self,
            signature: Optional[str] = None, version: Optional[float] = None,
            flags: Optional[int] = None, name: Optional[str] = None,
            central: Optional[Vector] = None, pivot: Optional[Vector] = None,
            bounding_box: Optional[BoundingBox] = None, material: Optional[str] = None,
            vertex_type: Optional[int] = None, indices: list[int] = [],
            positions: list[Vector] = [], uvs: list[Vector] = [],
            colors: list[list[int]] = []
        ):
        
        self.signature = signature
        self.version = version
        self.flags = flags
        self.name = name
        self.central = central
        self.pivot = pivot
        self.bounding_box = bounding_box
        self.material = material
        self.vertex_type = vertex_type
        self.indices = indices
        self.positions = positions
        self.uvs = uvs
        self.colors = colors

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

    def stream(self, path, mode, raw=None):
        if raw != None:
            if raw == True:  # the bool True value
                return BinStream(BytesIO())
            else:
                return BinStream(BytesIO(raw))
        return BinStream(open(path, mode))

    def stream_sco(self, path, mode, raw=None):
        if raw != None:
            if raw == True:  # the bool True value
                return StringIO()
            else:
                return StringIO(raw.decode('ascii'))
        return open(path, mode)

    def stream_scb(self, path, mode, raw=None):
        if raw != None:
            if raw == True:  # the bool True value
                return BinStream(BytesIO())
            else:
                return BinStream(BytesIO(raw))
        return BinStream(open(path, mode))

    def read_sco(self, path, raw=None):
        """
        Initialize the fields of :class:`SO`, using SCO.

        Parameters
        ------------
        `path`: Optional[:class:`str`]
            File path to read the SCO using an existing file that should work in `open(path)`.\n
        
        `raw`: Optional[:class:`bytes`]
            None (`Default value`): Recommended if reading from an file in local memory.\n
            bytes (`String encoded as ascii`): Reads SCO information from it.\n

        Raises
        --------
        `WrongFileSignature.`\n
        `InvalidMethodUsage.`\n
        """
        with self.stream_sco(path, 'r', raw) as f:
            lines = f.readlines()
            lines = [line[:-1] for line in lines]

            self.signature = lines[0]
            if self.signature != '[ObjectBegin]':
                raise Exception(
                    f'pyRitoFile: Failed: Read SCO {path}: Wrong file signature: {self.signature}')

            index = 1  # skip magic
            len1234 = len(lines)
            self.indices = []
            self.positions = []
            self.uvs = []
            while index < len1234:
                inp = lines[index].split()
                if len(inp) == 0:  # cant split, definitely not voldemort
                    index += 1
                    continue
                if inp[0] == 'Name=':
                    self.name = inp[1]

                elif inp[0] == 'CentralPoint=':
                    self.central = Vector(
                        float(inp[1]), float(inp[2]), float(inp[3]))

                elif inp[0] == 'PivotPoint=':
                    self.pivot = Vector(
                        float(inp[1]), float(inp[2]), float(inp[3]))

                elif inp[0] == 'Verts=':
                    vertex_count = int(inp[1])
                    for i in range(index+1, index+1 + vertex_count):
                        inp2 = lines[i].split()
                        self.positions.append(Vector(
                            float(inp2[0]), float(inp2[1]), float(inp2[2])))
                    index = i+1
                    continue

                elif inp[0] == 'Faces=':
                    face_count = int(inp[1])
                    for i in range(index+1, index+1 + face_count):
                        inp2 = lines[i].replace('\t', ' ').split()

                        # skip bad faces
                        face = (int(inp2[1]), int(inp2[2]), int(inp2[3]))
                        if face[0] == face[1] or face[1] == face[2] or face[2] == face[0]:
                            continue
                        self.indices.extend(face)

                        self.material = inp2[4]

                        # u v, u v, u v
                        self.uvs.append(
                            Vector(float(inp2[5]), float(inp2[6])))
                        self.uvs.append(
                            Vector(float(inp2[7]), float(inp2[8])))
                        self.uvs.append(
                            Vector(float(inp2[9]), float(inp2[10])))

                    index = i+1
                    continue

                index += 1

    def write_sco(self, path, raw=None):
        with self.stream_sco(path, 'w', raw) as f:
            f.write('[ObjectBegin]\n')  # magic
            f.write(f'Name= {self.name}\n') # name
            # central
            f.write(
                f'CentralPoint= {self.central.x:.4f} {self.central.y:.4f} {self.central.z:.4f}\n') \
            # pivot
            if self.pivot != None:
                f.write(
                    f'PivotPoint= {self.pivot.x:.4f} {self.pivot.y:.4f} {self.pivot.z:.4f}\n')
            # positions
            f.write(f'Verts= {len(self.positions)}\n')
            for position in self.positions:
                f.write(f'{position.x:.4f} {position.y:.4f} {position.z:.4f}\n')
            # faces
            face_count = len(self.indices) // 3
            f.write(f'Faces= {face_count}\n')
            for i in range(face_count):
                index = i * 3
                f.write('3\t')
                f.write(f' {self.indices[index]:>5}')
                f.write(f' {self.indices[index+1]:>5}')
                f.write(f' {self.indices[index+2]:>5}')
                f.write(f'\t{self.material:>20}\t')
                f.write(f'{self.uvs[index].x:.12f} {self.uvs[index].y:.12f} ')
                f.write(
                    f'{self.uvs[index+1].x:.12f} {self.uvs[index+1].y:.12f} ')
                f.write(
                    f'{self.uvs[index+2].x:.12f} {self.uvs[index+2].y:.12f}\n')
            f.write('[ObjectEnd]')
            return f.getvalue() if raw else None

    def read_scb(self, path, raw=None):
        """
        Initialize the fields of :class:`SO`, using a SCB.

        Parameters
        ------------
        `path`: Optional[:class:`str`]
            File path to read the SCB using an existing file that should work in `open(path)`.\n
        
        `raw`: Optional[:class:`bytes`]
            None (`Default value`): Recommended if reading from an file in local memory.
            bytes: Reads SCB information from a bytes object.\n

        Raises
        --------
        `WrongFileSignature.`\n
        `UnsupportedFileVersion.`\n
        `InvalidMethodUsage.`\n
        """
        with self.stream_scb(path, 'rb', raw) as bs:
            self.signature, = bs.read_s(8)
            if self.signature != 'r3d2Mesh':
                raise Exception(
                    f'pyRitoFile: Failed: Read SCB {path}: Wrong file signature: {self.signature}')

            major, minor = bs.read_u16(2)
            self.version = float(f'{major}.{minor}')
            if major not in (3, 2) and minor != 1:
                raise Exception(
                    f'pyRitoFile: Failed: Read SCB {path}: Unsupported file version: {major}.{minor}')

            self.name, = bs.read_s_padded(128)
            vertex_count, face_count, self.flags = bs.read_u32(3)
            self.flags = SOFlag(self.flags)

            # bouding box
            self.bounding_box = (bs.read_vec3()[0], bs.read_vec3()[0])

            if major == 3 and minor == 2:
                self.vertex_type, = bs.read_u32()

            self.positions = bs.read_vec3(vertex_count)
            if self.vertex_type != None:
                if self.vertex_type >= 1:
                    self.colors = [bs.read_u8(4) for i in range(vertex_count)]

            self.central, = bs.read_vec3()
            self.indices = []
            self.uvs = []
            for i in range(face_count):
                face = bs.read_u32(3)
                if face[0] == face[1] or face[1] == face[2] or face[2] == face[0]:
                    continue
                self.indices.extend(face)

                self.material, = bs.read_s_padded(64)

                uvs = bs.read_f32(6)

                # u u u, v v v
                self.uvs.append(Vector(uvs[0], uvs[3]))
                self.uvs.append(Vector(uvs[1], uvs[4]))
                self.uvs.append(Vector(uvs[2], uvs[5]))

    def write_scb(self, path, raw=None):
        def get_bounding_box():
            min = Vector(float("inf"), float("inf"), float("inf"))
            max = Vector(float("-inf"), float("-inf"), float("-inf"))
            for position in self.positions:
                if min.x > position.x:
                    min.x = position.x
                if min.y > position.y:
                    min.y = position.y
                if min.z > position.z:
                    min.z = position.z
                if max.x < position.x:
                    max.x = position.x
                if max.y < position.y:
                    max.y = position.y
                if max.z < position.z:
                    max.z = position.z
            return min, max
        
        with self.stream_scb(path, 'wb', raw) as bs:
            # signature, version
            bs.write_s('r3d2Mesh')  
            bs.write_u16(3, 2)  
            # name but nothing is fine
            bs.write_s_padded('', 128)  
            # position count, face count, flags
            face_count = len(self.indices) // 3
            bs.write_u32(len(self.positions), face_count, self.flags.value)
            bs.write_vec3(*get_bounding_box()) # bounding box
            bs.write_u32(0)  # vertex color
            # positions
            bs.write_vec3(*self.positions)
            # central
            bs.write_vec3(self.central)
            # faces 
            for i in range(face_count):
                index = i * 3
                bs.write_u32(
                    self.indices[index], self.indices[index+1], self.indices[index+2])

                bs.write_s_padded(self.material, 64)
                # u u u, v v v
                bs.write_f32(
                    self.uvs[index].x, self.uvs[index +
                                                1].x, self.uvs[index+2].x,
                    self.uvs[index].y, self.uvs[index+1].y, self.uvs[index+2].y
                )


from base64 import b64encode
from xxhash import xxh3_64_intdigest
from io import BytesIO
from .stream import BinStream
from json import (
    loads as json_loads,
    dumps as json_dumps
)


class RST:
    __slots__ = ('entries', 'version', 'hash_bits', 'has_trenc', 'count', 'hashtable')

    def __init__(self, entries = dict(), version = 5, hash_bits = 38, has_trenc=False, count = 0, hashtable = dict()):
        self.entries = entries
        self.version = version
        self.hash_bits = hash_bits
        self.has_trenc = has_trenc
        self.count = count
        self.hashtable = hashtable

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

    def stream(self, path, mode, raw=None):
        if raw != None:
            if raw == True:  # the bool True value
                return BinStream(BytesIO())
            else:
                return BinStream(BytesIO(raw))
        return BinStream(open(path, mode))

    def read(self, path, raw=None):
        # Credits https://github.com/CommunityDragon/CDTB/blob/master/cdtb/rstfile.py
        with self.stream(path, 'rb', raw) as f:
            magic, = f.read_s(3)
            self.version, = f.read_u8()
            if magic != "RST":
                raise Exception(f'pyRitoFile: invalid magic for RST file "{magic}"')
            
            if self.version != 5:
                raise Exception(f'pyRitoFile: unsupported RST version "{self.version}"')
            

            hash_mask = (1 << self.hash_bits) - 1
            self.count, = f.read_ul()
            entries = []

            # what is this bro ;-;
            for _ in range(self.count):
                v, = f.read_ull()
                entries.append((v >> self.hash_bits, v & hash_mask))
            
            data = f.stream.read()
            for offset, b_hash in entries:
                end = data.find(b"\0", offset)
                d = data[offset:end]
                self.entries[b_hash] = d.decode('utf-8', 'replace')
            
            print()

    def write(self, path, raw=None):
        with self.stream(path, 'wb', raw) as bs:
            bs.write_s("RST")
            bs.write_u8(self.version)
            bs.write_ul(len(self.entries))
            
            entries_order = []
            for b_hash_str, text in self.entries.items():
                try:
                    hash_int = int(b_hash_str)
                except:
                    hash_int = xxh3_64_intdigest(b_hash_str.lower()) & ((1 << self.hash_bits) - 1)
                
                entries_order.append((hash_int, text))
            
            data_block = bytearray()
            current_offset = 0
            entries_with_offsets = []
            
            for hash_int, text in entries_order:
                text_bytes = text.encode('utf-8', 'replace') + b'\x00'
                entries_with_offsets.append((hash_int, current_offset))
                data_block.extend(text_bytes)

                current_offset += len(text_bytes)
            
            hash_mask = (1 << self.hash_bits) - 1
            for hash_int, offset in entries_with_offsets:
                combined = (offset << self.hash_bits) | (hash_int & hash_mask)
                bs.write_ull(combined)
            
            bs.stream.write(data_block)

    def read_json(self, path, raw=None):
        if raw:
            self.entries = json_loads(raw)
        else:
            self.entries = json_loads(open(path, 'r', encoding='utf-8').read())

    def write_json(self, path, raw=None):
        result = json_dumps(self.entries, indent=4, ensure_ascii=False)
        if raw:
            return result

        with open(path, 'w', encoding='utf-8') as f:
            f.write(result)
    
    def un_hash(self):
        new_entries = {}
        for h, value in self.entries.items():
            if h in self.hashtable:
                new_entries[self.hashtable[h]] = value
            else:
                new_entries[int(h)] = value
        
        self.entries = new_entries


from io import BytesIO
from .stream import BinStream
from enum import IntEnum

class BNKHelper:
    @staticmethod
    def skip_init_params(bs):
        bs.pad(bs.read_u8()[0] * 5)
        bs.pad(bs.read_u8()[0] * 9)

    @staticmethod
    def skip_pos_params(bs, bkhd_version):
        pos_bits, = bs.read_u8()
        has_pos = pos_bits & 1
        has_3d = False
        has_automation = False
        if has_pos:
            if bkhd_version <= 89:
                has_2d, has_3d = bs.read_b(2)
                if has_2d: bs.pad(1)
            else:
                has_3d = pos_bits & 2
        if has_pos and has_3d:
            if bkhd_version <= 89:
                has_automation = (bs.read_u8()[0] & 3) != 1
                bs.pad(8)
            else:
                has_automation = (pos_bits >> 5) & 3
                bs.pad(1)
        if has_automation:
            bs.pad(9 if bkhd_version <= 89 else 5)
            bs.pad(16 * bs.read_u32()[0])
            bs.pad((16 if bkhd_version <= 89 else 20) * bs.read_u32()[0])
        elif bkhd_version <= 59:
            bs.pad(1)

    
    @staticmethod 
    def skip_aux(bs):
        has_aux = (bs.read_u8()[0] >> 3) & 1
        if has_aux: bs.pad(16)
        bs.pad(6)
        bs.pad(3 * bs.read_u8()[0])
        for i in range(bs.read_u8()[0]):
            bs.pad(5)
            bs.pad(8 * bs.read_u8()[0])
            
    @staticmethod
    def skip_rtpc(bs, bkhd_version):
        rtpc_count, = bs.read_u16()
        for i in range(rtpc_count):
            bs.pad(13 if bkhd_version <= 89 else 12)
            bs.pad(12 * bs.read_u16()[0])


    @staticmethod
    def skip_base_params(bs, bkhd_version):
        bs.pad(1)
        fx_count, = bs.read_u8()
        bs.pad(5 + int(fx_count != 0) - int(bkhd_version <= 89) + (fx_count * 7))

        parent_id, = bs.read_u32()
    
        bs.pad(2 if bkhd_version <= 89 else 1)
        BNKHelper.skip_init_params(bs)
        BNKHelper.skip_pos_params(bs, bkhd_version)
        BNKHelper.skip_aux(bs)
        BNKHelper.skip_rtpc(bs, bkhd_version)

        return parent_id
    
    @staticmethod
    def skip_clip_automation(bs):
        for i in range(bs.read_u32()[0]):
            bs.pad(8)
            bs.pad(12 * bs.read_u32()[0])
    

class BNKObjectType(IntEnum):
    Settings = 1
    Sound = 2
    Action = 3
    Event = 4
    RandomOrSequenceContainer = 5
    SwitchContainer = 6
    ActorMixer = 7
    AudioBus = 8
    BlendContainer = 9
    MusicSegment = 10
    MusicTrack = 11
    MusicSwitchContainer = 12
    MusicPlaylistContainer = 13
    Attenuation = 14
    DialogueEvent = 15
    MotionBus = 16
    MotionFX = 17
    Effect = 18
    AuxiliaryBus = 19

    def __json__(self):
        return self.name


class BNKObjectData:
    def __init__(self):
        pass

    def __json__(self):
        return vars(self)


class BNKObject:
    # hirc
    __slots__ = (
        'id', 'type', 'size', 'data'
    )

    def __init__(self, id=None, type=None, size=None, data=None):
        self.id = id
        self.type = type
        self.size = size
        self.data = data

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}


class BNKWem:
    # didx
    __slots__ = (
        'id', 'offset', 'size'
    )

    def __init__(self, id=None, offset=None, size=None):
        self.id = id
        self.offset = offset
        self.size = size

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}


class BNKSectionData:
    def __init__(self):
        pass

    def __json__(self):
        return vars(self)


class BNKSection:
    __slots__ = (
        'signature', 'size', 'data'
    )

    def __init__(self, signature=None, size=None, data=None):
        self.signature = signature
        self.size = size
        self.data = data

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}


class BNK:
    __slots__ = ('bkhd', 'didx', 'data', 'hirc', 'unknown_sections')

    def __init__(self, bkhd=None, didx=None, data=None, hirc=None, unknown_sections=None):
        self.bkhd = bkhd
        self.didx = didx
        self.data = data
        self.hirc = hirc
        self.unknown_sections = unknown_sections

    def __json__(self):
        return {key: getattr(self, key) for key in self.__slots__}

    def stream(self, path, mode, raw=None):
        if raw != None:
            if raw == True:  # the bool True value
                return BinStream(BytesIO())
            else:
                return BinStream(BytesIO(raw))
        return BinStream(open(path, mode))

    def read(self, path, raw=None):
        with self.stream(path, 'rb', raw) as bs:
            self.unknown_sections = []
            while bs.tell() < bs.end():
                section = BNKSection()
                section.signature, = bs.read_s(4)
                section.size, = bs.read_u32()
                section.data = BNKSectionData()
                if section.signature == 'BKHD':
                    # bank header: data is depend of version
                    self.bkhd = bkhd = section.data
                    bkhd.version, bkhd.id = bs.read_u32(2)
                    bkhd.unk = bs.read(section.size - 8)
                elif section.signature == 'DIDX':
                    # data index: contains list of wems(id, offset, size)
                    self.didx = didx = section.data
                    wem_count = section.size // 12
                    didx.wems = []
                    for i in range(wem_count):
                        wem = BNKWem()
                        wem.id, wem.offset, wem.size = bs.read_u32(3)
                        didx.wems.append(wem)
                elif section.signature == 'DATA':
                    # data: all wems data, should just save start offset
                    self.data = data = section.data
                    data.start_offset = bs.tell()
                    bs.pad(section.size)
                elif section.signature == 'HIRC':
                    # hierarchy: contains list of wwise objects
                    self.hirc = hirc = section.data
                    object_count, = bs.read_u32()
                    hirc.objects = []
                    for i in range(object_count):
                        obj = BNKObject()
                        obj.type = BNKObjectType(bs.read_u8()[0])
                        obj.size, obj.id = bs.read_u32(2)
                        obj.data = BNKObjectData()
                        obj_offset = bs.tell()
                        if obj.type == BNKObjectType.Sound:
                            sound = obj.data
                            bs.pad(4)
                            if self.bkhd.version == 88:
                                sound.stream_type, = bs.read_u32()
                            else:
                                sound.stream_type, = bs.read_u8()
                            sound.wem_id, sound.source_id, = bs.read_u32(
                                2)
                            if self.bkhd.version == 88:
                                bs.pad(7)
                            else:
                                bs.pad(8)
                            sound.object_id, = bs.read_u32()
                        elif obj.type == BNKObjectType.Action:
                            action = obj.data
                            action.scope, = bs.read_u8()
                            action.type, = bs.read_u8()
                            if action.type == 25:
                                bs.pad(5)
                                BNKHelper.skip_init_params(bs)
                                action.switch_group_id, action.switch_id = bs.read_u32(2)
                            else:
                                action.object_id, = bs.read_u32()       
                        elif obj.type == BNKObjectType.Event:
                            event = obj.data
                            if self.bkhd.version == 58:
                                action_id_count, = bs.read_u32()
                            else:
                                action_id_count, = bs.read_u8()
                            event.action_ids = bs.read_u32(action_id_count)
                        elif obj.type == BNKObjectType.RandomOrSequenceContainer:
                            container = obj.data
                            container.switch_container_id = BNKHelper.skip_base_params(bs, self.bkhd.version)
                            bs.pad(24)
                            container.sound_ids = bs.read_u32(bs.read_u32()[0])
                        elif obj.type in (BNKObjectType.MusicSegment, BNKObjectType.MusicPlaylistContainer):
                            segment = obj.data
                            bs.pad(4)
                            segment.music_switch_id, segment.sound_id = bs.read_u32(2)
                            bs.pad(1)
                            BNKHelper.skip_init_params(bs)
                            BNKHelper.skip_pos_params(bs, self.bkhd.version)
                            BNKHelper.skip_aux(bs)
                            BNKHelper.skip_rtpc(bs, self.bkhd.version)
                            segment.music_track_ids = bs.read_u32(bs.read_u32()[0])
                        elif obj.type == BNKObjectType.MusicTrack:
                            track = obj.data 
                            bs.pad(1)
                            bs.pad(14 * bs.read_u32()[0])
                            playlist_count, = bs.read_u32()
                            bs.pad(playlist_count * 44)
                            track.track_count, = bs.read_u32()
                            bs.pad(0 - 4 - playlist_count * 44)
                            
                            track.wem_ids = [0] * track.track_count
                            for i in range(playlist_count):
                                track_index, wem_id, event_id = bs.read_u32(3)
                                play_at, begin_strim_offset, end_trim_offset, source_duration = bs.read_f64(4)
                                track.wem_ids[track_index] = wem_id
                            bs.pad(4)
                            BNKHelper.skip_clip_automation(bs)
                            track.parent_id = BNKHelper.skip_base_params(bs, self.bkhd.version)
                            if bs.read_u8()[0] == 3:
                                track.has_switch_ids = True
                                bs.pad(1)
                                track.switch_group_id, = bs.read_u32()
                                bs.pad(4+4)
                                track.switch_ids = bs.read_u32(track.track_count)

                                
                        obj_size = bs.tell() - obj_offset + 4
                        if obj_size < obj.size:
                            bs.pad(obj.size-obj_size)
                        hirc.objects.append(obj)
                else:
                    # unknown sections to read
                    section.data = bs.read(section.size)
                    self.unknown_sections.append(section)


    def write(self, path, wem_datas, raw=None):
        with self.stream(path, 'wb', raw) as bs:
            # write bkhd
            # signature, size, version, id, unknown 24 bytes
            bs.write_s('BKHD')
            bs.write_u32(32, 134, 0)
            bs.write(b'>]p\x17\x00\x00\x00\x00\xfa\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00') 

            # write didx
            # signature, size
            wem_data_offsets = []
            bs.write_s('DIDX')
            bs.write_u32(len(self.didx.wems)*12)
            # write wem infos
            for i, wem in enumerate(self.didx.wems):
                wem_data_offsets.append(bs.tell()+4)
                wem.size = len(wem_datas[i])
                bs.write_u32(wem.id, 0, wem.size)

            # write data
            # signature, size
            bs.write_s('DATA')
            bs.write_u32(sum(wem.size for wem in self.didx.wems))
            # wem data - need to minus start offset
            start_offset = bs.tell()
            for i, wem_data in enumerate(wem_datas):
                data_offset = bs.tell()
                bs.seek(wem_data_offsets[i])
                bs.write_u32(data_offset-start_offset)
                bs.seek(data_offset)
                bs.write(wem_data)
                
            return bs.raw() if raw else None





def Elf(s):
    h = 0
    for c in s.lower():
        h = (h << 4) + ord(c)
        t = (h & 0xF0000000)
        if t != 0:
            h ^= (t >> 24)
        h &= ~t
    return h


def FNV1(s):
    h = 0x811c9dc5
    for b in s.encode('ascii').lower():
        h = ((h * 0x01000193) % 0x100000000) ^ b
    return h


def FNV1a(s):
    h = 0x811c9dc5
    for b in s.encode('ascii').lower():
        h = ((h ^ b) * 0x01000193) % 0x100000000
    return h
